##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################

"""
This module contains utility functions for use in testing IDAES models.
"""

__author__ = "Andrew Lee"


from pyomo.environ import Set, SolverFactory, Var
from pyomo.common.config import ConfigBlock

from idaes.core import (declare_process_block_class,
                        PhysicalParameterBlock,
                        StateBlock,
                        StateBlockData,
                        ReactionParameterBlock,
                        ReactionBlockBase,
                        ReactionBlockDataBase,
                        MaterialFlowBasis,
                        MaterialBalanceType,
                        EnergyBalanceType,
                        MomentumBalanceType)


def get_default_solver():
    """
    Tries to set-up the default solver for testing, and returns None if not
    available
    """
    if SolverFactory('ipopt').available(exception_flag=False):
        solver = SolverFactory('ipopt')
        solver.options = {'tol': 1e-6,
                          'linear_solver': 'ma27'}
    else:
        solver = None

    return solver


# -----------------------------------------------------------------------------
# Define some generic PhysicalBlock and ReactionBlock classes for testing
@declare_process_block_class("PhysicalParameterTestBlock")
class _PhysicalParameterBlock(PhysicalParameterBlock):
    def build(self):
        super(_PhysicalParameterBlock, self).build()

        self.phase_list = Set(initialize=["p1", "p2"])
        self.component_list = Set(initialize=["c1", "c2"])
        self.phase_equilibrium_idx = Set(initialize=["e1", "e2"])
        self.element_list = Set(initialize=["H", "He", "Li"])
        self.element_comp = {"c1": {"H": 1, "He": 2, "Li": 3},
                             "c2": {"H": 4, "He": 5, "Li": 6}}

        self.phase_equilibrium_list = \
            {"e1": ["c1", ("p1", "p2")],
             "e2": ["c2", ("p1", "p2")]}

        # Attribute to switch flow basis for testing
        self.basis_switch = 1
        self.default_balance_switch = 1

        self.state_block_class = TestStateBlock

    @classmethod
    def define_metadata(cls, obj):
        obj.add_default_units({'time': 's',
                               'length': 'm',
                               'mass': 'g',
                               'amount': 'mol',
                               'temperature': 'K',
                               'energy': 'J',
                               'holdup': 'mol'})


class SBlockBase(StateBlock):
    def initialize(blk, outlvl=0, optarg=None, solver=None,
                   hold_state=False, **state_args):
        for k in blk.keys():
            blk[k].init_test = True
            blk[k].hold_state = hold_state

    def release_state(blk, flags=None, outlvl=0):
        for k in blk.keys():
            blk[k].hold_state = not blk[k].hold_state


@declare_process_block_class("TestStateBlock", block_class=SBlockBase)
class StateTestBlockData(StateBlockData):
    CONFIG = ConfigBlock(implicit=True)

    def build(self):
        super(StateTestBlockData, self).build()

        self.flow_vol = Var(initialize=20)
        self.flow_mol_phase_comp = Var(self._params.phase_list,
                                       self._params.component_list,
                                       initialize=2)
        self.test_var = Var(initialize=1)
        self.pressure = Var(initialize=1e5)
        self.temperature = Var(initialize=300)

        self.enth_mol = Var(initialize=10000)

        self.gibbs_mol_phase_comp = Var(self._params.phase_list,
                                        self._params.component_list,
                                        initialize=50)
        self.entr_mol = Var(initialize=1000)

    def get_material_flow_terms(b, p, j):
        return b.test_var

    def get_material_density_terms(b, p, j):
        return b.test_var

    def get_enthalpy_flow_terms(b, p):
        return b.test_var

    def get_energy_density_terms(b, p):
        return b.test_var

    def model_check(self):
        self.check = True

    def get_material_flow_basis(b):
        if b.config.parameters.basis_switch == 1:
            return MaterialFlowBasis.molar
        elif b.config.parameters.basis_switch == 2:
            return MaterialFlowBasis.mass
        else:
            return MaterialFlowBasis.other

    def default_material_balance_type(self):
        if self._params.default_balance_switch == 1:
            return MaterialBalanceType.componentPhase
        else:
            raise NotImplementedError

    def default_energy_balance_type(self):
        if self._params.default_balance_switch == 1:
            return EnergyBalanceType.enthalpyTotal
        else:
            raise NotImplementedError

    def define_state_vars(self):
        return {"component_flow": self.flow_mol_phase_comp,
                "temperature": self.temperature,
                "pressure": self.pressure}


@declare_process_block_class("ReactionParameterTestBlock")
class _ReactionParameterBlock(ReactionParameterBlock):
    def build(self):
        super(_ReactionParameterBlock, self).build()

        self.phase_list = Set(initialize=["p1", "p2"])
        self.component_list = Set(initialize=["c1", "c2"])
        self.rate_reaction_idx = Set(initialize=["r1", "r2"])
        self.equilibrium_reaction_idx = Set(initialize=["e1", "e2"])

        self.rate_reaction_stoichiometry = {("r1", "p1", "c1"): 1,
                                            ("r1", "p1", "c2"): 1,
                                            ("r1", "p2", "c1"): 1,
                                            ("r1", "p2", "c2"): 1,
                                            ("r2", "p1", "c1"): 1,
                                            ("r2", "p1", "c2"): 1,
                                            ("r2", "p2", "c1"): 1,
                                            ("r2", "p2", "c2"): 1}
        self.equilibrium_reaction_stoichiometry = {
                                            ("e1", "p1", "c1"): 1,
                                            ("e1", "p1", "c2"): 1,
                                            ("e1", "p2", "c1"): 1,
                                            ("e1", "p2", "c2"): 1,
                                            ("e2", "p1", "c1"): 1,
                                            ("e2", "p1", "c2"): 1,
                                            ("e2", "p2", "c1"): 1,
                                            ("e2", "p2", "c2"): 1}

        self.reaction_block_class = ReactionBlock

        # Attribute to switch flow basis for testing
        self.basis_switch = 1

    @classmethod
    def define_metadata(cls, obj):
        obj.add_default_units({'time': 's',
                               'length': 'm',
                               'mass': 'g',
                               'amount': 'mol',
                               'temperature': 'K',
                               'energy': 'J',
                               'holdup': 'mol'})

    @classmethod
    def get_required_properties(self):
        return {}


class RBlockBase(ReactionBlockBase):
    def initialize(blk, outlvl=0, optarg=None, solver=None):
        for k in blk.keys():
            blk[k].init_test = True


@declare_process_block_class("ReactionBlock", block_class=RBlockBase)
class ReactionBlockData(ReactionBlockDataBase):
    CONFIG = ConfigBlock(implicit=True)

    def build(self):
        super(ReactionBlockData, self).build()

        self.reaction_rate = Var(["r1", "r2"])

        self.dh_rxn = {"r1": 10,
                       "r2": 20,
                       "e1": 30,
                       "e2": 40}

    def model_check(self):
        self.check = True

    def get_reaction_rate_basis(b):
        if b.config.parameters.basis_switch == 1:
            return MaterialFlowBasis.molar
        elif b.config.parameters.basis_switch == 2:
            return MaterialFlowBasis.mass
        else:
            return MaterialFlowBasis.other



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
GAMS writer for the regression
"""
import sys
from . import (
    DataImport,
    BasisFunctions,
    GAMSDataWrite,
    DataManipulation,
    #SoaveDensity
)

import numpy as np
import matplotlib.pyplot as plt

global props, sample, sample_ratio
props = []
sample = False
sample_ratio = 5

global num_terms
molecule, data_name, runFile = "", "", ""
constraints = 0
Combination = False
max_time = 1000
num_terms = 12
Rm = 8.314472
critT, critP, critD, M, triple, acc = [0, 0, 0, 0, 0, 0]


# General
def molData(Dfluids, Dmolecule, Ddata_name, Dterms, Dmax_time):
    """ Passing of the Data from the main module ::module:: MPEOSDeveloperModule

        :param Dmolecule: Name of molecule of interest
        :type Dmolecule: str
        :param Ddata_name: Name of Data files
        :type Ddata_name: str.
        :param Dmax_time: Running time limit for the GAMS file.
        :type Dmax_time: int.
    """
    global molecule, data_name, max_time, constraints
    global critT, critP, critD, M, triple, acc
    global num_terms
    (critT, critP, critD, M, triple, acc) = Dfluids

    molecule = Dmolecule
    max_time = Dmax_time
    num_terms = Dterms
    data_name = Ddata_name
    # print "Constraints", Dconstraints
    # constraints = Dconstraints;


def openFile(data_name, ending=".gms"):
    """ Opens the GAMS file based on the data set

        :param data_name: name of the molecule
        :type data_name: str
    """
    global textFile
    name = molecule + data_name + ending
    print("Creating file: %s" % name)
    try:
        textFile = open(name, "w")
    except Exception:
        print("couldn't open the file Something wrong")
        sys.exit(0)


def closeFile():
    """ Closes the GAMS file"""
    textFile.close()
    print("Closing File")


def getTextFile():
    "Returns name of the current textFile"
    global textFile
    return textFile


def setCombination(isCombination):
    "Sets the regression as multiple properties"
    global Combination
    Combination = isCombination


def getRunFile():
    "Returns the gams file name"
    global runFile
    return runFile


def setNumberTerms(numterms):
    "Set number of basis functions allowed in regression"
    global num_terms
    num_terms = numterms


# Data Import
def importData():
    global props
    global sample, sample_ratio

    if "PVT" in props:
        DataImport.PVT(molecule, sample=sample)
    # DataImport.samplePVT6(molecule); #TOL
    if "CV" in props:
        DataImport.CV(molecule)

    if "CP" in props:
        DataImport.CP(molecule, sample=sample)
        # DataImport.sampleCP1(molecule);
    # DataImport.sampleCP2(molecule);
    # DataImport.sampleCP3(molecule);  #TOL
    if "SND" in props:
        DataImport.SND(molecule, sample=sample)


# GDX file
def GenerateGDXGamsFiledtlmv():
    """
    Generates a gams file that creates a data .gdx file
    Creates Combination of PVT, CV, CP, and SND GAMS file. (Titles precoded in).
    """
    global Combination, data_name, props

    Combination = True
    LemJac = False
    BasisFunctions.formCustomBasis(LemJac)
    terms = len(BasisFunctions.coeffs)

    openFile(runFile)

    indexes = {}
    PVTValues, CVValues, CPValues, SNDValues = [], [], [], []
    dataPoints = 0
    if "PVT" in props:
        PVTValues = DataImport.PVTValues
        indexes["PVT"] = DataImport.PVTindexes

    if "CV" in props:
        CVValues = DataImport.CVValues
        indexes["CV"] = DataImport.CVindexes

    if "CP" in props:
        CPValues = DataImport.CPValues
        indexes["CP"] = DataImport.CPindexes

    if "SND" in props:
        SNDValues = DataImport.SNDValues
        indexes["SND"] = DataImport.SNDindexes

    props = indexes.keys()
    dataPoints = [len(PVTValues), len(CVValues), len(CPValues), len(SNDValues)]

    writeGamsHeaderdtl(dataPoints, terms, "", ["PVT", "CV", "CP", "SND"], indexes)

    textFile.write("$Offlisting\n")

    GAMSDataWrite.writeExp(textFile)

    DataToWrite = []
    if "PVT" in props:
        for x in PVTValues:
            manVal = DataManipulation.PVT(x)
            DataToWrite.append(manVal)
        GAMSDataWrite.PVTdt(textFile, DataToWrite, Combination)

    if "CV" in props:
        DataToWrite = []
        for x in CVValues:
            manVal = DataManipulation.CV(x)
            DataToWrite.append(manVal)
        GAMSDataWrite.CVdt(textFile, DataToWrite, Combination)

    if "CP" in props:
        DataToWrite = []
        for x in CPValues:
            manVal = DataManipulation.CP(x)
            DataToWrite.append(manVal)
        GAMSDataWrite.CPdt(textFile, DataToWrite, Combination)

    if "SND" in props:
        DataToWrite = []
        for x in SNDValues:
            manVal = DataManipulation.SND(x)
            DataToWrite.append(manVal)
        GAMSDataWrite.SNDdt(textFile, DataToWrite, Combination)

    textFile.write("$Onlisting\n")

    listProps = ""
    for p in props:
        listProps = "%s %s," % (listProps, p)
        for val, r in zip(indexes[p], ["G", "L", "C", "LD", "MD", "HD"]):
            if val > 0:
                listProps = "%s %s%s," % (listProps, p, r)

    if DataImport.isothermIndex != 0:
        listProps = "%s IT," % (listProps)

    textFile.write(
        "execute_unload '%sData.gdx', p, i, j, itt, delta,tau,d,t,l, %s z;\n"
        % (molecule, listProps)
    )


# GDX Header
def writeGamsHeaderdtl(num_points, terms, kset, pset, regions=None):
    """ Writes the GAMS file Header including the number of terms and data points as well as different thermodynamic properties.

        :param num_points: number of data points
        :type num_points: int
        :param terms: number of basis functions
        :type terms: int
        :param kset: what is this
        :param pset: what is this

        .. todo:: What is the kset and pset?
    """
    # print num_points
    global Combination
    global props
    # print "Combination", Combination
    textFile.write("$offdigit\n$offsymxref offsymlist\n")
    if len(num_points) >= 1:
        textFile.write("set i /1 * %d/" % max(num_points))
        for data, prop in zip(num_points, pset):
            if prop in props:
                textFile.write("\n %s(i) /1 * %d/" % (prop, data))
                if regions[prop] is not None:
                    prevx = 0
                    # print regions[prop]
                    for x, r in zip(regions[prop], ["G", "L", "C", "LD", "MD", "HD"]):
                        if x > 1:
                            if r == "HD":
                                textFile.write(
                                    "\n %s%s(i) /%d * %d/" % (prop, r, prevx + 1, x)
                                )
                            else:
                                textFile.write(
                                    "\n %s%s(i) /%d * %d/" % (prop, r, prevx + 1, x)
                                )
                            prevx = x
                        elif x >0:
                            if r == "HD":
                                textFile.write(
                                    "\n %s%s(i) /%d/" % (prop, r, x)
                                )
                            else:
                                textFile.write(
                                    "\n %s%s(i) /%d/" % (prop, r, x)
                                )
                            prevx = x
                    if DataImport.isothermIndex != 0 and prop == "PVT":
                        textFile.write(
                            "\n %s(i) /%d * %d/" % ("IT", 1, DataImport.isothermIndex)
                        )
                    if len(DataImport.InSatValues) > 0 and prop == "PVT":  # 5/9
                        textFile.write(
                            "\n %s(i) /%d * %d/"
                            % ("IS", 1, len(DataImport.InSatValues))
                        )
        textFile.write(";\n")
    else:
        textFile.write("set i /1 * %d/" % max(num_points))
        prevx = 0
        if prop in props and regions is not None:
            for x, r in zip(regions, ["G", "L", "C", "LD", "MD", "HD"]):
                if x > 0:
                    # print x
                    # print regions
                    if r == "HD":
                        textFile.write("\n %s(i) /%d * %d/" % (r, prevx + 1, x))
                    else:
                        textFile.write("\n %s(i) /%d * %d/" % (r, prevx + 1, x))
                    prevx = x
            if DataImport.isothermIndex != 0:
                textFile.write(
                    "\n %s(i) /%d * %d/" % ("IT", 1, DataImport.isothermIndex)
                )
        textFile.write(";\n")

    textFile.write("set j /1* %d/" % terms)
    prevx = 1
    textFile.write(";\n")

    if Combination:
        propString = "Crit"
        for p in pset:
            if propString == "":
                propString = propString + "%s" % p
            else:
                propString = propString + ", %s" % p
        textFile.write("set p /%s/;\n" % propString)
        # textFile.write("parameters z(p, i), xijk(p, i,j, k),crit(p,i,j,k), isoT(p,i,j,k);\n parameters betalo(j), betaup(j);\n")
        textFile.write(
            "parameters z(p, i), d(j), l(j), t(j), delta(p,i), tau(p,i), itt(p,i);\n"
        ) 
    else:
        textFile.write(
            "parameters z(i),  d(j), l(j), t(j), delta(i), tau(i), itt(i);\n"
        ) 


# GAMS Shells
def GenerateGamsShell():
    """
    Generates the multiparameter equation of state regression through gams. 
    Imports the molecule data.gdx file
    Creates Combination of PVT, CV, CP, and SND GAMS file. (Title precoded in).
    """
    global Combination
    global data_name
    global props

    Combination = True
    BasisFunctions.formCustomBasis()
    terms = len(BasisFunctions.coeffs)

    openFile(runFile)

    indexes = {}
    PVTValues, CVValues, CPValues, SNDValues = [], [], [], []

    if "PVT" in props:
        PVTValues = DataImport.PVTValues
        indexes["PVT"] = DataImport.PVTindexes

    if "CV" in props:
        CVValues = DataImport.CVValues
        indexes["CV"] = DataImport.CVindexes

    if "CP" in props:
        CPValues = DataImport.CPValues
        indexes["CP"] = DataImport.CPindexes

    if "SND" in props:
        SNDValues = DataImport.SNDValues
        indexes["SND"] = DataImport.SNDindexes

    writeGamsShellHeaderB(props, indexes)

    DataToWrite = []
    print(indexes["SND"])

    PVTVar = []
    if "PVT" in props:
        for x in PVTValues:
            manVal = DataManipulation.PVT(x)
            DataToWrite.append(manVal)
        last = 0
        PVTVar = []
        for y in indexes["PVT"]:
            if not y == 0:
                Z = [x[2] for x in DataToWrite[last:y]]
                if len(Z) > 1:
                    AvgZ = sum(Z) / len(Z)
                    Z_1 = [(AvgZ - x) ** 2 for x in Z]
                    VarZ = sum(Z_1) / len(Z)
                    last = y
                    PVTVar.append(VarZ)
                else:
                    PVTVar.append(1)
            else:
                PVTVar.append(0)

    DataToWrite = []
    CVVar = []
    if "CV" in props:
        for x in CVValues:
            manVal = DataManipulation.CV(x)
            DataToWrite.append(manVal)

        last = 0
        for y in indexes["CV"]:
            if not y == 0:
                CV = [x[2] for x in DataToWrite[last:y]]
                AvgCV = sum(CV) / len(CV)
                CV_1 = [(AvgCV - x) ** 2 for x in CV]
                VarCV = sum(CV_1) / len(CV)
                CVaad = [x for x in CV]
                VarCV = np.var(CVaad)
                last = y
                CVVar.append(VarCV)
            else:
                CVVar.append(0)

    DataToWrite = []
    CPVar = []
    if "CP" in props:
        for x in CPValues:
            manVal = DataManipulation.CP(x)
            DataToWrite.append(manVal)

        last = 0
        for y in indexes["CP"]:
            if not y == 0:
                CP = [x[2] for x in DataToWrite[last:y]]
                VarCP = 1
                if len(CP) > 1:
                    AvgCP = sum(CP) / len(CP)
                    CP_1 = [(AvgCP - x) ** 2 for x in CP]
                    VarCP = sum(CP_1) / len(CP)
                last = y
                if VarCP == 0:
                    VarCP = 1
                CPVar.append(VarCP)
            else:
                CPVar.append(0)

    DataToWrite = []
    SNDVar = []
    if "SND" in props:
        for x in SNDValues:
            manVal = DataManipulation.SND(x)
            DataToWrite.append(manVal)

        SNDVar = []
        last = 0
        for y in indexes["SND"]:
            if not y == 0:
                W = [x[2] for x in DataToWrite[last:y]]
                if len(W) > 1:
                    AvgW = sum(W) / len(W)
                    W_1 = [(AvgW - x) ** 2 for x in W]
                    VarW = sum(W_1) / len(W)
                last = y
                if VarW < 1e-5:
                    VarW = 1
                SNDVar.append(VarW)
            else:
                SNDVar.append(0)
        print(SNDVar)

    # Write Variances
    textFile.write("Parameter variance(p,i);\n")
    for var, r in zip(PVTVar, ["G", "L", "C", "LD", "MD", "HD"]):
        if var > 0:
            textFile.write("variance('PVT',PVT%s) = %f;\n" % (r, var))
    for var, r in zip(CVVar, ["G", "L", "C", "LD", "MD", "HD"]):
        if var > 0:
            textFile.write("variance('CV',CV%s) = %f;\n" % (r, var))
    for var, r in zip(CPVar, ["G", "L", "C", "LD", "MD", "HD"]):
        if var > 0:
            textFile.write("variance('CP',CP%s) = %f;\n" % (r, var))
    for var, r in zip(SNDVar, ["G", "L", "C", "LD", "MD", "HD"]):
        if var > 0:
            textFile.write("variance('SND',SND%s) = %f;\n" % (r, var))

    textFile.write("$Offlisting\n")
    writeConstants(terms)
    textFile.write("$Onlisting\n")

    writeDerivatives(props)
    writeBasisFunctions()

    writeEquationsAndVariablesB(props)

    textFile.write("\n\n******** OBJECTIVES\n\n")
    if "PVT" in props and len(props) > 1:
        writeObjectivesB(["PVT"])
    writeObjectivesB(props)
    writeBoundsB(props)

    writeGamsShellFooterB(data_name, False)

    textFile.write("\n\n******** SOLVES\n\n")
    if "PVT" in props and len(props) > 1:
        writeModelB(500, ["PVT"])
    writeCalculateIntermediates(props)
    writeModelB(3600, props)

    writeModelPostEvaluations(props)

    closeFile()


# GAMS Shell imports in GDX
def writeGamsShellHeaderB(pset, regions=None):
    """ Writes the GAMS file Header including the number of terms and data points as well as different thermodynamic properties.

        :param pset: list of thermodynamic properties
        :param regions: regions of thermodynamic property data
    """
    textFile.write("$offdigit\n$offsymxref offsymlist\n")

    textFile.write("**************** IMPORT\n")
    textFile.write("Set jl /0*398/\n j(jl) /1*398/\n jld(j) /72*96/;\n\n")

    textFile.write("$Gdxin %sData\n" % molecule)

    setText = "Sets i"
    loadSetText = "$load i"
    for prop in pset:
        if regions[prop] is not None:
            setText = "%s, %s(i)" % (setText, prop)
            loadSetText = "%s, %s" % (loadSetText, prop)
            for x, r in zip(regions[prop], ["G", "L", "C", "LD", "MD", "HD"]):
                if x > 0:
                    setText = "%s, %s%s(%s)" % (setText, prop, r, prop)
                    loadSetText = "%s, %s%s" % (loadSetText, prop, r)

            if DataImport.isothermIndex != 0 and prop == "PVT":
                setText = "%s, IT(i)" % (setText)
                loadSetText = "%s, IT" % (loadSetText)
            # if len(DataImport.InSatValues)>0 and prop =="PVT": # 5/9
            #   setText = "%s, IS(i)"%(setText)
            #   loadSetText = "%s, IS"%(loadSetText);
    textFile.write("%s;\n" % setText)
    textFile.write("%s\n\n" % loadSetText)

    textFile.write("Sets p /'Crit', 'PVT', 'CV', 'CP', 'SND'/;")

    textFile.write("\n\n**************** IMPORT BASIS FUNCTION VALUES\n")
    textFile.write(
        "parameters z(p,i), delta(p,i), tau(p,i), itt(p,i), d(j), l(j), t(j);\n"
    )
    textFile.write("$loaddc z delta tau itt d l t\n")
    textFile.write("$gdxin\n\n")

    textFile.write("Set Crit(i) /1/;\ndelta('Crit','1') = 1;\ntau('Crit','1') = 1;\n\n")

    textFile.write("Parameter data(p,i);\n")
    for prop in pset:
        textFile.write("data('%s', %s(i)) = 1;\n" % (prop, prop))
    textFile.write("data('Crit',Crit(i)) =1;\n\n")


def writeConstants(terms):
    """ Writes down the ranges of the fitting Beta value and sets regular to one.
        param terms: number of basis functions
        param terms: int
    """
    textFile.write("Parameter betalo(j), betaup(j);\n")
    betalo = -5
    betaup = 5
    i = "j"
    textFile.write("betalo(%s) = %d ;" % (i, betalo))
    textFile.write("betaup(%s) =  %d ;\n" % (i, betaup))


def writeGamsShellFooterB(data_name, load_in=False):
    """
        Writes the Gam Footer options, model, and display.
    """

    textFile.write("\n\n******** GENERAL OPTIONS\n")

    textFile.write("option optca = 0.0001; \n")
    textFile.write("option optcr = 0.0001 ;\n")
    textFile.write("option limrow = 0, limcol =0 ;\n")
    textFile.write("option solprint=off;\n")
    textFile.write("option reslim = %s;\n" % max_time)
    textFile.write("option sys12 = 1;\n")
    textFile.write("maxexecerror = 10000;\n")

    textFile.write("option rminlp = baron;\n")
    textFile.write("option minlp = baron;\n")
    textFile.write("option decimals = 6;\n")
    textFile.write("option savepoint = 1;\n")


# EQUATIONS
def writeDerivatives(props):
    """
    Calculate and write the important derivatives of the basis functions
    """
    textFile.write("\n\n****** Derivatives\n\n")

    if "PVT" in props:
        textFile.write("parameter PVTvec(i,j), PVTvecCV(i,j);\n")
        textFile.write("PVTvec(PVT(i),j)$(l(j) = 0) = d(j);\n")
        textFile.write(
            "PVTvec(PVT(i),j)$(l(j) <> 0) = d(j)- l(j)*(power(delta('PVT',i),l(j)));\n\n"
        )

        textFile.write("PVTvecCV(PVT(i),j) = t(j)*(t(j)-1);\n")

    if "CV" in props:
        textFile.write("parameter CVvec(i,jl);\n")
        textFile.write("CVvec(CV(i),'0') = itt('CV',i);\n")
        textFile.write("CVvec(CV(i),j) = -t(j)*(t(j)-1);\n\n")

    if "CP" in props:
        textFile.write("parameter CPvecA(i,jl), CPvecB(i,jl), CPvecCV(i,jl);\n\n")

        textFile.write("CPvecA(CP(i),'0')= 1;\n")
        textFile.write("CPvecA(CP(i),j)$(l(j) =0) = 2*d(j) + d(j)*(d(j)-1);\n")
        textFile.write(
            "CPvecA(CP(i),j)$(l(j) <> 0) = 2*(d(j)-l(j)*power(delta('CP',i),l(j))) + (d(j)-l(j)*power(delta('CP',i),l(j)))*(d(j) - 1 - l(j)*power(delta('CP',i),l(j))) - power(l(j),2)*power(delta('CP',i),l(j));\n\n"
        )

        textFile.write("CPvecB(CP(i),'0') = 1;\n")
        textFile.write("CPvecB(CP(i),j)$(l(j) = 0) = d(j)*(1-t(j));\n")
        textFile.write(
            "CPvecB(CP(i),j)$(l(j) <> 0) = (d(j) - l(j)*power(delta('CP',i),l(j)))*(1-t(j));\n\n"
        )

        textFile.write("CPvecCV(CP(i),'0') = 0;\n")
        textFile.write("CPvecCV(CP(i),j) = -t(j)*(t(j)-1);\n")

    if "SND" in props:

        textFile.write("parameter SNDvecA(i,jl), SNDvecB(i,jl), SNDvecCV(i,jl);\n\n")

        textFile.write("SNDvecA(SND(i),'0')= 1;\n")
        textFile.write("SNDvecA(SND(i),j)$(l(j) =0) = 2*d(j) + d(j)*(d(j)-1);\n")
        textFile.write(
            "SNDvecA(SND(i),j)$(l(j) <> 0) = 2*(d(j)-l(j)*power(delta('SND',i),l(j))) + (d(j)-l(j)*power(delta('SND',i),l(j)))*(d(j) - 1 - l(j)*power(delta('SND',i),l(j))) - power(l(j),2)*power(delta('SND',i),l(j));\n\n"
        )

        textFile.write("SNDvecB(SND(i),'0') = 1;\n")
        textFile.write("SNDvecB(SND(i),j)$(l(j) = 0) = d(j)*(1-t(j));\n")
        textFile.write(
            "SNDvecB(SND(i),j)$(l(j) <> 0) = (d(j) - l(j)*power(delta('SND',i),l(j)))*(1-t(j));\n\n"
        )

        textFile.write("SNDvecCV(SND(i),'0') = -itt('SND',i);\n")
        textFile.write("SNDvecCV(SND(i),j) = -t(j)*(t(j)-1);\n\n")

    Z = (float(critP) * 1000 / Rm / float(critT) / float(critD)) - 1.00
    textFile.write("parameter CRITvec(jl) ,CRITvec1(jl), CRITvec2(jl);\n\n")
    textFile.write("CRITvec('0') = %f;\n" % Z)
    textFile.write("CRITvec(j)$(l(j) = 0) = -d(j);\n")
    textFile.write(
        "CRITvec(j)$(l(j) <> 0) = -d(j) + l(j)*power(delta('Crit','1'),l(j));\n\n"
    )

    textFile.write("CRITvec1('0')= 1;\n")
    textFile.write("CRITvec1(j)$(l(j) =0) = d(j)*(d(j)+1);\n")
    textFile.write(
        "CRITvec1(j)$(l(j) <> 0) = d(j)*(d(j)+1) - 2*(d(j)+1)*l(j)*power(delta('Crit','1'),l(j)) - l(j)*(l(j)+1)*power(delta('Crit','1'),l(j)) + power(l(j),2)*power(delta('Crit','1'),(2*l(j)));\n\n"
    )

    textFile.write("CRITvec2('0') = 0;\n")
    textFile.write(
        "CRITvec2(j)$(l(j)=0) = 2*d(j)*(2*d(j)-1) + d(j)*(d(j)-1)*(d(j)-2);\n"
    )
    textFile.write(
        "CRITvec2(j)$(l(j)<>0) = 2*(d(j)-l(j)) + 4*((d(j)-l(j))*(d(j)-1-l(j)) - power(l(j),2)) + (d(j)*(d(j)-1)*(d(j)-2) + (-2*l(j) + 6*d(j)*l(j) - 3 * power(d(j),2)*l(j) - 3*d(j)*power(l(j),2) + 3*power(l(j),2) - power(l(j),3)) + (3*d(j)*power(l(j),2) - 3*power(l(j),2) + 3*power(l(j),3))-power(l(j),3));\n\n"
    )

    textFile.write("parameter ITvec(i,j);\n")
    textFile.write("ITvec(i,j)$(l(j) = 0) = d(j)*(d(j)-1)*(d(j)-2);\n")
    textFile.write(
        "ITvec(i,j)$(l(j) <> 0 ) =d(j)*(d(j)-1)*(d(j)-2) + power(delta('PVT',i),l(j))*(-2*l(j)+6*d(j)*l(j)-3*(d(j)**2)*l(j)-3*d(j)*(l(j)**2)+3*(l(j)**2) - (l(j)**3)) + power(delta('PVT',i),2*l(j))*(3*d(j)*(l(j)**2) - 3*(l(j)**2) + 3*(l(j)**3)) - (l(j)**3)*power(delta('PVT',i),3*l(j));\n\n"
    )


def writeBasisFunctions():
    """
    Write the values of the basis function terms
    """
    textFile.write("\n\n********** BASIS FUNCTIONS\n\n")

    textFile.write("parameter base(p,i,jl);\n\n")

    textFile.write("base(p,i,'0') = 1;\n")
    textFile.write(
        "base(p, i,j)$(data(p,i)=1 and l(j) <> 0) = (delta(p,i)**d(j))* (tau(p,i)**t(j)) * exp(- (delta(p,i)**l(j)));\n"
    )
    textFile.write(
        "base(p,i,j)$(data(p,i)=1 and l(j) = 0) = (delta(p,i)**d(j))* (tau(p,i)**t(j));\n\n"
    )


def writeEquationsAndVariablesB(props):
    """
        Writes multiple thermodynamic parameter equations and constants.
        :param props: Array containing the available properties.
        :type props: array
    """
    global constraints
    # properties = len(props);

    textFile.write("\n\n********  EQUATIONS and VARIABLES\n\n")

    textFile.write("binary variable y(j);\n")
    textFile.write("variables beta(jl), sse;\n")

    eqtnList = ""
    posVars = ""
    Vars = "gammaCrit, gammaCrit1, gammaCrit2"
    if "PVT" in props:
        eqtnList = "%s gammaAPVT," % eqtnList
        Vars = "%s, gammaPVT(i)" % Vars
        if DataImport.isothermIndex != 0:
            eqtnList = "%s ITgamma," % eqtnList
            Vars = "%s,  gammaIT(i)" % Vars
    if "CV" in props:
        eqtnList = "%s gammaACV," % eqtnList
        if posVars == "":
            posVars = "%s gammaCV(i)" % posVars
        else:
            posVars = "%s, gammaCV(i)" % posVars
    if "CP" in props:
        eqtnList = "%s gammaACP, gammaBCP, gammaCCP, " % eqtnList
        Vars = "%s, gammaCPB(i), gammaCPC(i)" % Vars
        if posVars == "":
            posVars = "%s gammaCPA(i)" % posVars
        else:
            posVars = "%s, gammaCPA(i)" % posVars
    if "SND" in props:
        eqtnList = "%s gammaASND, gammaBSND, gammaCSND, omegaSND, psiSND" % eqtnList
        Vars = "%s, gammaSNDA(i), gammaSNDB(i), SNDomega(i)" % Vars
        if posVars == "":
            posVars = "%s gammaSNDC(i), SNDpsi(i)" % posVars
        else:
            posVars = "%s, gammaSNDC(i), SNDpsi(i)" % posVars

    textFile.write("variables %s;\n" % Vars)
    textFile.write("positive variables %s;\n" % posVars)
    textFile.write("beta.lo(j) = betalo(j); beta.up(j) = betaup(j);\n")
    textFile.write(
        "equations eq1, eq2, eq3, ITgamma, Critgamma, Critgamma1, Critgamma2,  \n  %s \n obj;\n"
        % eqtnList
    )
    # eq5

    textFile.write("beta.FX('0')=1;\n")
    textFile.write("eq1(j).. betalo(j)*y(j) =l= beta(j);\n")
    textFile.write("eq2(j).. beta(j) =l= betaup(j)*y(j);\n")
    textFile.write("eq3.. sum(j, y(j)) =e= %d;\n\n" % num_terms)

    # Write property equations.
    if "PVT" in props:
        textFile.write(
            "gammaAPVT(PVT(i)).. gammaPVT(i) =e= sum(j,beta(j)*PVTvec(i,j)*base('PVT',i,j));\n\n"
        )
        if DataImport.isothermIndex != 0:
            textFile.write(
                "ITgamma(IT(i)).. gammaIT(i) =e= sum(j, beta(j)*ITvec(i,j)*base('PVT',i,j));\n"
            )
    if "CV" in props:
        textFile.write(
            "gammaACV(CV(i)).. gammaCV(i) =e= sum(jl, beta(jl)*CVvec(i,jl)*base('CV',i,jl));\n\n"
        )
    if "CP" in props:
        textFile.write(
            "gammaACP(CP(i)).. gammaCPA(i) =e= sum(j,beta(j)*CPvecCV(i,j)*base('CP',i,j));\n"
        )
        textFile.write(
            "gammaBCP(CP(i)).. gammaCPB(i) =e= sum(jl,  beta(jl)*CPvecB(i,jl)*base('CP',i,jl));\n"
        )
        textFile.write(
            "gammaCCP(CP(i)).. gammaCPC(i) =e= sum(jl,beta(jl)*CPvecA(i,jl)*base('CP',i,jl));\n\n"
        )
        # textFile.write("omegaCP(CP(i)).. CPomega(i) =e= gammaCPA(i)*gammaCPC(i);\n")
        # textFile.write("psiCP(CP(i)).. CPpsi(i) =e= gammaCPB(i)*gammaCPB(i);\n\n")
    if "SND" in props:
        textFile.write(
            "gammaASND(SND(i)).. gammaSNDA(i) =e= sum(jl,beta(jl)*SNDvecA(i,jl)*base('SND',i,jl));\n"
        )
        textFile.write(
            "gammaBSND(SND(i)).. gammaSNDB(i) =e= sum(jl,beta(jl)*SNDvecB(i,jl)*base('SND',i,jl));\n"
        )
        textFile.write(
            "gammaCSND(SND(i)).. gammaSNDC(i) =e= sum(jl,beta(jl)*SNDvecCV(i,jl)*base('SND',i,jl));\n"
        )
        textFile.write(
            "omegaSND(SND(i)).. SNDomega(i) =e= gammaSNDA(i)*gammaSNDC(i);\n"
        )
        textFile.write(
            "psiSND(SND(i))..   SNDpsi(i) =e= gammaSNDB(i)*gammaSNDB(i);\n\n"
        )

    textFile.write("Critgamma..  gammaCrit =e= sum(jl, beta(jl)*Critvec(jl));\n")
    textFile.write("Critgamma1.. gammaCrit1 =e= sum(jl, beta(jl)*Critvec1(jl));\n")
    textFile.write("Critgamma2.. gammaCrit2 =e= sum(jl, beta(jl)*Critvec2(jl));\n\n")


def writeObjectivesB(props):
    """
    Writes the objective equations for the different combination of properites and constraints
    """
    propsString = ""
    for p in props:
        propsString = "%s%s" % (propsString, p)

    objString = "obj%s.. sse =e= " % propsString

    textFile.write("EQUATION obj%s;\n" % propsString)

    if "PVT" in props:
        objString = (
            "%s + sum(PVT(i),power(z('PVT',i) - gammaPVT(i),2)/variance('PVT',i))"
            % objString
        )
    if "CV" in props:
        objString = (
            "%s + sum(CV(i), power( z('CV',i) - gammaCV(i),2)/variance('CV',i))"
            % objString
        )
    if "CP" in props:
        objString = (
            "%s + sum(CP(i),power(z('CP',i) - gammaCPA(i) - power(gammaCPB(i),2)/gammaCPC(i),2)/variance('CP',i))"
            % objString
        )
    if "SND" in props:
        objString = (
            "%s + sum(SND(i),power(z('SND',i)*gammaSNDC(i) - SNDomega(i) - SNDpsi(i) ,2)/variance('SND',i))"
            % objString
        )
    if DataImport.isothermIndex != 0:
        objString = "%s + 100*sum(IT(i),power(gammaIT(i),2)) " % objString

    textFile.write(
        "%s  + 5* power( gammaCrit,2) + 5 * power(gammaCrit1,2)+ 5* power(gammaCrit2,2);\n\n"
        % objString
    )

    textFile.write("\n\n")


def writeBoundsB(props):
    """
    Writes bounds on the variables
    """
    textFile.write("\n\n****** Property Bounds\n\n")

    # textFile.write("y.fx(jld) = 0;\n")
    textFile.write("beta.up(j) = 5;\n")
    textFile.write("beta.lo(j) = -5;\n\n")

    if "PVT" in props:
        PVTVals = [DataManipulation.PVT(x)[2] for x in DataImport.PVTValues]
        minZ = np.min(PVTVals)
        maxZ = np.max(PVTVals)
        print("Min", minZ, "Max", maxZ)

        textFile.write("gammaPVT.up(PVT)$(z('PVT',PVT)>0) = %f;\n" % (maxZ * 10))
        textFile.write("gammaPVT.lo(PVT)$(z('PVT',PVT)>0) = %f;\n\n" % (-1))
        # textFile.write("gammaPVT.up(PVT)$(z('PVT',PVT)<0) = 0;\n");
        # textFile.write("gammaPVT.lo(PVT)$(z('PVT',PVT)<0) = -1;\n\n");

    if "CV" in props:
        textFile.write("*gammaCV.up(i) = z('CV',i)*1.05;\n")
        textFile.write("*gammaCV.lo(i) = z('CV',i)*0.95;\n\n")

    if "CP" in props:
        textFile.write("*gammaCPA.lo(i) = 0;\n")
        textFile.write("*gammaCPA.up(i) = 20;\n")
        textFile.write("*gammaCPB.lo(i) = 0;\n")
        textFile.write("*gammaCPB.up(i) = 20;\n")
        textFile.write("*gammaCPC.lo(i) = 1E-16;\n")
        textFile.write("*gammaCPC.up(i) = 40;\n\n")

    if "SND" in props:
        textFile.write("*gammaSNDA.up(i) = 50;\n")
        textFile.write("*gammaSNDA.lo(i) = 0;\n")
        textFile.write("*gammaSNDB.up(i) = 20;\n")
        textFile.write("*gammaSNDB.lo(i) = 0;\n")
        textFile.write("*gammaSNDC.up(i) = 20;\n")
        textFile.write("*gammaSNDC.lo(i) = 0;\n\n")

        textFile.write("*SNDomega.lo(i) = 0;\n")
        textFile.write("*SNDomega.up(i)  = 120;\n")
        textFile.write("*SNDpsi.lo(i) = 0;\n")
        textFile.write("*SNDpsi.up(i) = 100;\n\n")


def writeModelB(reslim, props):
    """
    Writes the GAMS model definition
    """
    propsString = ""
    for p in props:
        propsString = "%s%s" % (propsString, p)

    textFile.write("option reslim = %i;\n" % reslim)
    modelName = "sse_%s" % propsString
    # modelString = "model sse_%s /"%propsString;

    eqtnList = "eq1, eq2, eq3, Critgamma, Critgamma1, Critgamma2, obj%s" % propsString
    if "PVT" in props:
        eqtnList = "%s, gammaAPVT" % eqtnList
    if "CV" in props:
        eqtnList = "%s, gammaACV" % eqtnList
    if "CP" in props:
        eqtnList = "%s, gammaACP, gammaBCP, gammaCCP" % eqtnList
    if "SND" in props:
        eqtnList = "%s, gammaASND, gammaBSND, gammaCSND" % eqtnList
    if DataImport.isothermIndex != 0:
        eqtnList = "%s, ITgamma " % eqtnList

    modelString = "model %s / %s/;\n" % (modelName, eqtnList)
    solveString = "solve %s minimizing sse using minlp;\n" % (modelName)
    displayString = "display beta.l, y.l;\n\n"

    textFile.write(modelString)
    textFile.write(solveString)
    textFile.write(displayString)


def writeCalculateIntermediates(props):
    """
    Calculate intermediate values of properties not in the regression
    """

    # Write property equations.
    if "PVT" in props:
        textFile.write(
            "gammaPVT.l(i) = sum(j,beta.l(j)*PVTvec(i,j)*base('PVT',i,j));\n\n"
        )
        if DataImport.isothermIndex != 0:
            textFile.write(
                "gammaIT.l(i) = sum(j, beta.l(j)*ITvec(i,j)*base('PVT',i,j));\n"
            )
    if "CV" in props:
        textFile.write(
            "gammaCV.l(i) = sum(jl, beta.l(jl)*CVvec(i,jl)*base('CV',i,jl));\n\n"
        )
    if "CP" in props:
        textFile.write(
            "gammaCPA.l(i) = sum(j,beta.l(j)*CPvecCV(i,j)*base('CP',i,j));\n"
        )
        textFile.write(
            "gammaCPB.l(i) = sum(jl,  beta.l(jl)*CPvecB(i,jl)*base('CP',i,jl));\n"
        )
        textFile.write(
            "gammaCPC.l(i) = sum(jl,beta.l(jl)*CPvecA(i,jl)*base('CP',i,jl));\n\n"
        )
        # textFile.write("omegaCP(CP(i)).. CPomega(i) =e= gammaCPA(i)*gammaCPC(i);\n")
        # textFile.write("psiCP(CP(i)).. CPpsi(i) =e= gammaCPB(i)*gammaCPB(i);\n\n")
    if "SND" in props:
        textFile.write(
            "gammaSNDA.l(i) = sum(jl,beta.l(jl)*SNDvecA(i,jl)*base('SND',i,jl));\n"
        )
        textFile.write(
            "gammaSNDB.l(i) = sum(jl,beta.l(jl)*SNDvecB(i,jl)*base('SND',i,jl));\n"
        )
        textFile.write(
            "gammaSNDC.l(i) = sum(jl,beta.l(jl)*SNDvecCV(i,jl)*base('SND',i,jl));\n"
        )
        textFile.write(
            "SNDomega.l(i) = gammaSNDA.l(i)*gammaSNDC.l(i);\n"
        )
        textFile.write(
            "SNDpsi.l(i) = gammaSNDB.l(i)*gammaSNDB.l(i);\n\n"
        )

    # textFile.write("Critgamma..  gammaCrit =e= sum(jl, beta(jl)*Critvec(jl));\n")
    # textFile.write("Critgamma1.. gammaCrit1 =e= sum(jl, beta(jl)*Critvec1(jl));\n")
    # textFile.write("Critgamma2.. gammaCrit2 =e= sum(jl, beta(jl)*Critvec2(jl));\n\n")


def writeModelPostEvaluations(props):
    """
    Calculates thermodynamic values after the regression
    """

    textFile.write("\n\n****************Post Evaluation\n\n")

    textFile.write("display gammaCrit.l, gammaCrit1.l, gammaCrit2.l;\n")

    if DataImport.isothermIndex != 0:
        textFile.write("display gammaIT.l;\n")

    if "PVT" in props:
        textFile.write("display gammaPVT.l;\n")
    if "CV" in props:
        textFile.write("display gammaCV.l;\n")

    if "CP" in props:
        textFile.write("display gammaCPA.l, gammaCPB.l, gammaCPC.l;\n")

        textFile.write("parameter mingammaCPA, mingammaCPB, mingammaCPC;\n")
        textFile.write("mingammaCPA = smin(CP(i), gammaCPA.l(i));\n")
        textFile.write("mingammaCPB = smin(CP(i), gammaCPB.l(i));\n")
        textFile.write("mingammaCPC = smin(CP(i), gammaCPC.l(i));\n")
        textFile.write("display mingammaCPA, mingammaCPB, mingammaCPC;\n\n")

        textFile.write("parameter maxgammaCPA, maxgammaCPB, maxgammaCPC;\n")
        textFile.write("maxgammaCPA = smax(CP(i), gammaCPA.l(i));\n")
        textFile.write("maxgammaCPB = smax(CP(i), gammaCPB.l(i));\n")
        textFile.write("maxgammaCPC = smax(CP(i), gammaCPC.l(i));\n")
        # textFile.write("maxCPomega = smax(CP(i), CPomega.l(i));\n");
        # textFile.write("maxCPpsi = smax(CP(i), CPpsi.l(i));\n");
        textFile.write("display maxgammaCPA, maxgammaCPB, maxgammaCPC;\n\n")

    if "SND" in props:
        textFile.write("display gammaSNDA.l, gammaSNDB.l, gammaSNDC.l;\n")

        textFile.write(
            "parameter mingammaSNDA, mingammaSNDB, mingammaSNDC, minSNDomega, minSNDpsi;\n"
        )
        textFile.write("mingammaSNDA = smin(SND(i), gammaSNDA.l(i));\n")
        textFile.write("mingammaSNDB = smin(SND(i), gammaSNDB.l(i));\n")
        textFile.write("mingammaSNDC = smin(SND(i), gammaSNDC.l(i));\n")
        textFile.write("minSNDomega = smin(SND(i), SNDomega.l(i));\n")
        textFile.write("minSNDpsi = smin(SND(i), SNDpsi.l(i));\n")
        textFile.write(
            "display mingammaSNDA, mingammaSNDB, mingammaSNDC, minSNDomega, minSNDpsi;\n\n"
        )

        textFile.write(
            "parameter maxgammaSNDA, maxgammaSNDB, maxgammaSNDC, maxSNDomega, maxSNDpsi;\n"
        )
        textFile.write("maxgammaSNDA = smax(SND(i), gammaSNDA.l(i));\n")
        textFile.write("maxgammaSNDB = smax(SND(i), gammaSNDB.l(i));\n")
        textFile.write("maxgammaSNDC = smax(SND(i), gammaSNDC.l(i));\n")
        textFile.write("maxSNDomega = smax(SND(i), SNDomega.l(i));\n")
        textFile.write("maxSNDpsi = smax(SND(i), SNDpsi.l(i));\n")
        textFile.write(
            "display maxgammaSNDA, maxgammaSNDB, maxgammaSNDC, maxSNDomega, maxSNDpsi;\n\n"
        )

    errString = ""
    TotErrString = ""
    for p in props:
        errString = "%s err%s," % (errString, p)
        TotErrString = "%s + err%s" % (TotErrString, p)
    errString = "%s totErr;\n" % errString
    textFile.write("parameter %s" % errString)

    if "PVT" in props:
        textFile.write(
            "errPVT = sum(PVT,power(z('PVT',PVT) - gammaPVT.l(PVT),2)/variance('PVT',PVT));\n"
        )
    if "CV" in props:
        textFile.write(
            "errCV =  sum(CV(i), power( z('CV',i) - gammaCV.l(i),2)/variance('CV',i));\n"
        )
    if "CP" in props:
        textFile.write(
            "errCP =  sum(CP(i),power(z('CP',i) - gammaCPA.l(i) - power(gammaCPB.l(i),2)/gammaCPC.l(i),2)/variance('CP',i));\n"
        )
    if "SND" in props:
        textFile.write(
            "errSND=  sum(SND(i),power(z('SND',i)*gammaSNDC.l(i) - SNDomega.l(i) - SNDpsi.l(i) ,2)/variance('SND',i));\n"
        )
    if DataImport.isothermIndex != 0:
        TotErrString = "%s + 100*sum(IT(i),power(gammaIT.l(i),2)) " % TotErrString

    textFile.write(
        "totErr = %s  + 5* power( gammaCrit.l,2) + 5 * power(gammaCrit1.l,2)+ 5* power(gammaCrit2.l,2);\n"
        % TotErrString
    )

    textFile.write("display %s" % errString)



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################

from importlib import reload
cvalsim = None  # could cause problems


def allcard(xdata, zdata, xval, zval, **kwargs):
    # enumerate all model cardinalities via ccmiqp and 
    # use validation/cross-validaiton to determine 
    from idaes.surrogate import alamopy
    from alamopy.writethis import writethis
    import numpy as np
    import math
    # import sympy
    import random
    from random import shuffle
    # from scipy.optimize import curve_fit #2.7
    # from scipy.optimize import minimize #2.7
    from scipy.linalg import lstsq  # 2.7
    import os
    # from alamopy import almlsq as almfun
    # from alamopy import almlsqjac as almjac
    import time
    # from sympy.parsing.sympy_parser import parse_expr
    import sys

    random.seed(100)

    # kwargs=opdict
    trans = list(['linfcns', 'expfcns', 'logfcns', 'sinfcns', 'cosfcns', 
                  'monomialpower', 'multi2power', 'multi3power', 'ratiopower'])
    et = list(['multi2power', 'multi3power', 'ratiopower'])
    # datacc = {}
    ins = list(['cvfold'])
    for opt in ins:
        if (opt in kwargs.keys()):
            cvfold = kwargs['cvfold']
    ntrans = 0
    ndata = np.shape(xdata)[0]
    ninputs = np.shape(xdata)[1]
    
    for t in trans:
        if t in kwargs.keys():
            if (type(kwargs[t]) == list()):
                nt = len(kwargs[t])
            else:
                nt = 1
            if t not in et:
                ntrans = ntrans + ninputs * nt
            else:
                ntrans = ntrans + math.factorial(ninputs) * nt
        else:
            kwargs[t] = 0
    # mseold = 0.0
    rmseold = 0.0
    startt = time.time()
    oldres = {}
    oldp = ()
    ntrans = min(ntrans, 1000)

    # split training and validationd ata before looping through cc
    tlist = list([])
    vlist = list([])
    # ndlist = range(ndata)
    if (cvfold == 'loo'):
        for i in range(ndata):
            vlist.append(np.asarray([i]))
            temp = [x for x in range(ndata) if x != i]
            tlist.append(np.asarray(temp))
    else:
        temp = range(ndata)
        shuffle(temp)
        # tlist = np.asarray(temp)
        if (cvfold == 'valset'):
            tlist = np.asarray(temp)[0:int(0.7 * ndata)]
            vlist = np.asarray(temp)[int(0.7 * ndata):-1]
        else:
            vlist = np.array_split(np.asarray(temp), int(cvfold))
            tlist = [1] * int(cvfold)
            for v in range(len(vlist)):
                tlist[v] = range(ndata)
                for this in vlist[v]:
                    tlist[v].remove(this)

    for ccon in list(range(1, ntrans + 1)):
        # reload(alamopy)
        # try:
        #     del cvalsim, almsim
        # except Exception:
        #     pass

        # res = alamopy.doalamo(xdata,zdata,kwargs.values())
        if (cvfold != 'valset'):
            res = alamopy.doalamo(xdata, zdata, xval=xval, zval=zval,  
                                  linfcns=kwargs['linfcns'],  
                                  expfcns=kwargs['expfcns'],  
                                  logfcns=kwargs['logfcns'],  
                                  sinfcns=kwargs['sinfcns'],  
                                  cosfcns=kwargs['cosfcns'], 
                                  monomialpower=kwargs['monomialpower'], 
                                  multi2power=kwargs['multi2power'],  
                                  multi3power=kwargs['multi3power'], 
                                  ratiopower=kwargs['ratiopower'], 
                                  sigma=kwargs['sigma'], 
                                  xlabels=kwargs['xlabels'], 
                                  zlabels=kwargs['zlabels'], 
                                  modeler=6, convpen=0, maxterms=ccon, 
                                  almname=kwargs['almname'], 
                                  expandoutput=kwargs['expandoutput'], 
                                  xmax=kwargs['xmax'], xmin=kwargs['xmin'], 
                                  savescratch=kwargs['savescratch'])
        else:
            res = alamopy.doalamo(xdata[tlist, :], zdata[tlist], 
                                  xval=xdata[vlist, :], zval=zdata[vlist], 
                                  linfcns=kwargs['linfcns'], 
                                  expfcns=kwargs['expfcns'], 
                                  logfcns=kwargs['logfcns'], 
                                  sinfcns=kwargs['sinfcns'], 
                                  cosfcns=kwargs['cosfcns'], 
                                  monomialpower=kwargs['monomialpower'], 
                                  multi2power=kwargs['multi2power'], 
                                  multi3power=kwargs['multi3power'], 
                                  ratiopower=kwargs['ratiopower'], 
                                  sigma=kwargs['sigma'], 
                                  xlabels=kwargs['xlabels'], 
                                  zlabels=kwargs['zlabels'], 
                                  modeler=6, convpen=0, maxterms=ccon, 
                                  almname=kwargs['almname'], 
                                  expandoutput=kwargs['expandoutput'], 
                                  xmax=kwargs['xmax'], xmin=kwargs['xmin'],
                                  savescratch=kwargs['savescratch'])
        for fn in ['cvalsim.py', 'cvalsim.pyc', 'almsim.py', 'almsim.pyc']:
            try:
                os.remove(fn)
            except Exception:
                pass
        os.system('cp ' + kwargs['almname'].split('.')[0] + 'cv.py cvalsim.py')
        os.system('cp ' + kwargs['almname'].split('.')[0] + 'alm.py almsim.py')

        if (ccon == 1):
            if (ndata < 10):  # not enough data to do cross validation
                writethis('Not enough data to facilitate cross validation')
                endt = time.time()
                res['totaltime'] = endt - startt
                return res

        # import cvalsim
        import almsim
        xalm = alamopy.almfeatmat(xdata, ccon)
        # Lets do the cross validatione error
        mseval = 0.0
        rmse = {}
        if (cvfold == 'valset'):
            resid = np.sum((zdata[vlist] - almsim.f(xdata[vlist, :]))**2)
            mseval = resid / float(len(vlist))
            params = 'ALM params used for valset'
            rmse = {}
            rmse['val'] = res['rmseval']
            rmse['train'] = res['rmse']
        else:
            mseval = 0.0
            # track = 0
            for tl, vl in zip(tlist, vlist):
                # initb = [ 0.0 for x in range(ccon)]
                xd, xmax, xmin = alamopy.mapminmax(xalm)
                zd, zmax, zmin = alamopy.mapminmax(zdata)
                fitres = lstsq(xd[tl], zd[tl])
                params = fitres[0]
                resid = zd[vl] - np.matmul(xd[vl, :], params[:])
                resid = alamopy.remapminmax(resid, zmax, zmin)
                resid = sum(resid ** 2)
                if (cvfold == 'loo'):
                    mseval = mseval + resid
                else:
                    mseval = mseval + resid / float(len(vl))
            mseval = mseval / float(len(vlist))
            rmse['val'] = np.sqrt(mseval)
            rmse['train'] = 'not'
        if (ccon > 1):
            if (float(rmse['val']) >= float(rmseold)):
                sys.stdout.write('              Problem name   : ' 
                                 + kwargs['almname'] + '\n')
                sys.stdout.write('  rMSEval : ' + str(rmse['val']) 
                                 + '    MSEold : ' + str(rmseold) + '\n')
                if (cvfold == 'valset'):
                    sys.stdout.write('   Ntrain : ' + str(len(tlist)) 
                                     + '    Nval : ' + str(len(vlist)) + '\n')
                else:
                    sys.stdout.write('   Ntrain : ' + str(len(tlist[0])) 
                                     + '    Nval : ' + str(len(vlist[0])) + '\n')
                sys.stdout.write('       optimal model size is : ' 
                                 + str(ccon - 1) + '\n')
                sys.stdout.write('    optimal coefficients are : ' 
                                 + str(oldp) + '\n')
                os.remove(kwargs['almname'].split('.')[0] + 'alm.py')
                os.system('cp ' + 'oldalmsim.sv ' + kwargs['almname'].split('.')[0] 
                          + 'alm.py')
                endt = time.time()
                oldres['totaltime'] = endt - startt
                return oldres
            elif (ccon == ntrans):
                endt = time.time()
                sys.stdout.write('optimal model size is :' + str(ccon) + '\n')
                res['totaltime'] = endt - startt
                return res
            else:
                # mseold = mseval
                oldres = res
        else:
            # mseold = float(mseval)
            rmseold = float(rmse['val'])
            oldres = res
            oldp = params
        # keep track of alm model of old iteratoin
        try:
            os.remove('oldalmsim.sv')
        except Exception:
            pass
        os.system('mv almsim.py oldalmsim.sv')


def almlsq(params, *X):
    # return lsq objective
    # import numpy as np
    try:
        reload(cvalsim)
    except Exception:
        # import cvalsim
        from cvalsim import f as sim

    xdata, zdata = X[0]    
    return sum((zdata - sim(xdata, params))**2)


def almlsqjac(params, *X):
    try:
        reload(cvalsim)
    except Exception:
        # import cvalsim
        from cvalsim import f as sim
    import numpy as np

    xdata, zdata = X[0]
    rp = np.ones(len(params))
    for i in range(len(params)):
        dparams = [0.0 for x in range(len(params))]
        dparams[i] = 1.0
        rp[i] = sum(-2.0 * (zdata - sim(xdata, params)) * sim(xdata, dparams))
    return rp
    # return np.dot(params,np.matmul(np.transpose(xdata),xdata))+np.matmul(zdata,xdata)


def almfeatmat(X, nparams):
    import numpy as np
    try:
        reload(cvalsim)
    except Exception:
        # import cvalsim
        from cvalsim import f as sim

    dims = np.shape(X)
    fm = np.ones([dims[0], nparams])
    dparams = [0.0 for x in range(nparams)]
    for i in range(nparams):
        dparams[i] = 1.0
        fm[:, i] = sim(X, dparams)
        dparams[i] = 0.0
    return fm



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################

"""
This module contains miscellaneous utility functions for use in IDAES models.
"""
import xml.dom.minidom
import pyomo.environ as pyo

# Author: Andrew Lee
def add_object_reference(self, local_name, remote_object):
    """
    Method to create a reference in the local model to a remote Pyomo object.
    This method should only be used where Pyomo Reference objects are not
    suitable (such as for referencing scalar Pyomo objects where the None
    index is undesirable).

    Args:
        local_name : name to use for local reference (str)
        remote_object : object to make a reference to

    Returns:
        None
    """
    try:
        object.__setattr__(self, local_name, remote_object)
    except AttributeError:
        raise AttributeError(
            "{} failed to construct reference to {} - remote "
            "object does not exist.".format(self.name, remote_object)
        )


# Author: Jaffer Ghouse
def extract_data(data_dict):
    """
    General method that returns a rule to extract data from a python
    dictionary. This method allows the param block to have a database for
    a parameter but extract a subset of this data to initialize a Pyomo
    param object.
    """

    def _rule_initialize(m, *args):
        if len(args) > 1:
            return data_dict[args]
        else:
            return data_dict[args[0]]

    return _rule_initialize


# Author: John Eslick
def TagReference(s, description=""):
    """
    Create a Pyomo reference with an added description string attribute to
    describe the reference. The intended use for these references is to create a
    time-indexed reference to variables in a model corresponding to plant
    measurment tags.

    Args:
        s: Pyomo time slice of a variable or expression
        description (str): A description the measurment

    Returns:
        A Pyomo Reference object with an added doc attribute
    """
    r = pyo.Reference(s)
    r.description = description
    return r


# Author John Eslick
def svg_tag(
    tags,
    svg,
    outfile=None,
    idx=None,
    tag_map=None,
    show_tags=False,
    byte_encoding="utf-8",
):
    """
    Replace text in a SVG with tag values for the model. This works by looking
    for text elements in the SVG with IDs that match the tags or are in tag_map.

    Args:
        tags: A dictionary where the key is the tag and the value is a Pyomo
            Refernce.  The refernce could be indexed. In yypical IDAES
            applications the references would be indexed by time.
        svg: a file pointer or a string continaing svg contents
        outfile: a file name to save the results, if None don't save
        idx: if None not indexed, otherwise an index in the indexing set of the
            reference
        tag_map: dictionary with svg id keys and tag values, to map svg ids to
            tags
        show_tags: Put tag labels of the diagram instead of numbers
        byte_encoding: If svg is given as a byte-array, use this encoding to
            convert it to a string.

    Returns:
        String for SVG
    """
    if isinstance(svg, str):  # assume this is svg content string
        pass
    elif isinstance(svg, bytes):
        svg = svg.decode(byte_encoding)
    elif hasattr(svg, "read"):  # file-like object to svg
        svg = svg.read()
    else:
        raise TypeError("SVG must either be a string or a file-like object")
    # Make tag map here because the tags may not make valid XML IDs if no
    # tag_map provided we'll go ahead and handle XML @ (maybe more in future)
    if tag_map is None:
        tag_map = dict()
        for tag in tags:
            new_tag = tag.replace("@", "_")
            tag_map[new_tag] = tag
    # Search for text in the svg that has an id in tags
    doc = xml.dom.minidom.parseString(svg)
    texts = doc.getElementsByTagName("text")
    for t in texts:
        id = t.attributes["id"].value
        if id in tag_map:
            # if it's multiline change last line
            tspan = t.getElementsByTagName("tspan")[-1].childNodes[0]
            try:
                if show_tags:
                    val = tag_map[id]
                elif idx is None:
                    val = pyo.value(tags[tag_map[id]], exception=False)
                else:
                    val = pyo.value(tags[tag_map[id]][idx], exception=False)
            except ZeroDivisionError:
                val = "Divide_by_0"
            try:
                tspan.nodeValue = "{:.4e}".format(val)
            except ValueError:  # whatever it is can't be scientific notation
                tspan.nodeValue = val

    new_svg = doc.toxml()
    # If outfile is provided save to a file
    if outfile is not None:
        with open(outfile, "w") as f:
            f.write(new_svg)
    return new_svg


# Author: John Eslick
def copy_port_values(destination, source):
    """
    Copy the variable values in the source port to the destination port. The
    ports must containt the same variables.

    Args:
        (pyomo.Port): Copy values from this port
        (pyomo.Port): Copy values to this port

    Returns:
        None
    """
    for k, v in destination.vars.items():
        if isinstance(v, pyo.Var):
            for i in v:
                v[i].value = pyo.value(source.vars[k][i])



# -*- coding: UTF-8 -*-
##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
This module contains utility functions for initialization of IDAES models.
"""

from pyomo.environ import Block, Var
from pyomo.network import Arc

from idaes.core.util.exceptions import ConfigurationError

__author__ = "Andrew Lee, John Siirola"


def fix_state_vars(blk, state_args={}):
    """
    Method for fixing state variables within StateBlocks. Method takes an
    optional argument of values to use when fixing variables.

    Args:
        blk : An IDAES StateBlock object in which to fix the state variables
        state_args : a dict containing values to use when fixing state
                variables. Keys must match with names used in the
                define_state_vars method, and indices of any variables must
                agree.

    Returns:
        A dict keyed by block index, state variable name (as defined by
        define_state_variables) and variable index indicating the fixed status
        of each variable before the fix_state_vars method was applied.
    """
    # For sanity, handle cases where state_args is None
    if state_args is None:
        state_args = {}

    flags = {}
    for k in blk.keys():
        for n, v in blk[k].define_state_vars().items():
            for i in v:
                flags[k, n, i] = v[i].is_fixed()

                # If not fixed, fix at either guess provided or current value
                if not v[i].is_fixed():
                    if n in state_args:
                        # Try to get initial guess from state_args
                        try:
                            if i is None:
                                val = state_args[n]
                            else:
                                val = state_args[n][i]
                        except KeyError:
                            raise ConfigurationError(
                                'Indexes in state_args did not agree with '
                                'those of state variable {}. Please ensure '
                                'that indexes for initial guesses are correct.'
                                .format(n))
                        v[i].fix(val)
                    else:
                        # No guess, try to use current value
                        if v[i].value is not None:
                            v[i].fix()
                        else:
                            # No initial value - raise Exception before this
                            # gets to a solver.
                            raise ConfigurationError(
                                'State variable {} does not have a value '
                                'assigned. This usually occurs when a Var '
                                'is not assigned an initial value when it is '
                                'created. Please ensure all variables have '
                                'valid values before fixing them.'
                                .format(v.name))

    return flags


def revert_state_vars(blk, flags):
    """
    Method to revert the fixed state of the state variables within an IDAES
    StateBlock based on a set of flags of the previous state.

    Args:
        blk : an IDAES StateBlock
        flags : a dict of bools indicating previous state with keys in the form
                (StateBlock index, state variable name (as defined by
                define_state_vars), var indices).

    Returns:
        None
    """
    for k in blk.keys():
        for n, v in blk[k].define_state_vars().items():
            for i in v:
                try:
                    if not flags[k, n, i]:
                        v[i].unfix()
                except KeyError:
                    raise ConfigurationError(
                        'Indices of flags proved do not match with indices of'
                        'the StateBlock. Please make sure you are using the '
                        'correct StateBlock.')


def propagate_state(stream, direction="forward"):
    """
    This method propagates values between Ports along Arcs. Values can be
    propagated in either direction using the direction argument.

    Args:
        stream : Arc object along which to propagate values
        direction: direction in which to propagate values. Default = 'forward'
                Valid value: 'forward', 'backward'.

    Returns:
        None
    """
    if not isinstance(stream, Arc):
        raise TypeError("Unexpected type of stream argument. Value must be "
                        "a Pyomo Arc.")

    if direction == "forward":
        value_source = stream.source
        value_dest = stream.destination
    elif direction == "backward":
        value_source = stream.destination
        value_dest = stream.source
    else:
        raise ValueError("Unexpected value for direction argument: ({}). "
                         "Value must be either 'forward' or 'backward'."
                         .format(direction))

    for v in value_source.vars:
        for i in value_source.vars[v]:
            if not isinstance(value_dest.vars[v], Var):
                raise TypeError("Port contains one or more members which are "
                                "not Vars. propogate_state works by assigning "
                                "to the value attribute, thus can only be "
                                "when Port members are Pyomo Vars.")
            if not value_dest.vars[v][i].fixed:
                value_dest.vars[v][i].value = value_source.vars[v][i].value


# HACK, courtesy of J. Siirola
def solve_indexed_blocks(solver, blocks, **kwds):
    """
    This method allows for solving of Indexed Block components as if they were
    a single Block. A temporary Block object is created which is populated with
    the contents of the objects in the blocks argument and then solved.

    Args:
        solver : a Pyomo solver object to use when solving the Indexed Block
        blocks : an object which inherits from Block, or a list of Blocks
        kwds : a dict of argumnets to be passed to the solver

    Returns:
        A Pyomo solver results object
    """
    # Check blocks argument, and convert to a list of Blocks
    if isinstance(blocks, Block):
        blocks = [blocks]

    try:
        # Create a temporary Block
        tmp = Block(concrete=True)

        nBlocks = len(blocks)

        # Iterate over indexed objects
        for i, b in enumerate(blocks):
            # Check that object is a Block
            if not isinstance(b, Block):
                raise TypeError("Trying to apply solve_indexed_blocks to "
                                "object containing non-Block objects")
            # Append components of BlockData to temporary Block
            try:
                tmp._decl["block_%s" % i] = i
                tmp._decl_order.append((b, i+1 if i < nBlocks-1 else None))
            except Exception:
                raise Exception("solve_indexed_blocks method failed adding "
                                "components to temporary block.")

        # Set ctypes on temporary Block
        tmp._ctypes[Block] = [0, nBlocks-1, nBlocks]

        # Solve temporary Block
        results = solver.solve(tmp, **kwds)

    finally:
        # Clean up temporary Block contents so they are not removed when Block
        # is garbage collected.
        tmp._decl = {}
        tmp._decl_order = []
        tmp._ctypes = {}

    # Return results
    return results



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################


def almconfidence(data, *vargs):
    # This function calulates a covariance matrix 
    # and confidence intervals of the estimated alamo regression coeficients
    import numpy as np
    # import sympy
    from scipy.stats import t  # 2.7
    from sympy.parsing.sympy_parser import parse_expr
    from sympy import symbols, lambdify

    if('xdata' not in data.keys()):
        xdata = vargs[0]
        # zdata = vargs[1]
    else:
        xdata = data['xdata']
        # zdata = data['zdata']
        
    ndata = np.shape(xdata)[0]
    # ninputs = np.shape(xdata)[1]
    # noutputs = np.shape(zdata)[1]
    if(isinstance(data['model'], type({}))):
        for okey in data['model'].keys():
            model = data['model'][okey]
            model.split('=')[1]
            # out = model.split('=')[0]
            model = model.split('=')[1]
            # split the model on +/- to isolate each linear term
            # This section is not currently in compliance with custom basis functions
            model = model.replace(' - ', ' + ').split(" + ")
            nlinterms = len(model)
            sensmat = np.zeros([ndata, nlinterms])
            covar = np.zeros([nlinterms, nlinterms])
            coeffs = np.zeros([nlinterms])
            for j in range(nlinterms):
                thisterm = model[j].split(' * ')
                coeffs[j] = float(eval(thisterm[0]))
                thisterm = thisterm[-1]
                thislam = lambdify([symbols(data['xlabels'])], 
                                   parse_expr(thisterm.replace('^', '**')), "numpy")
                for i in range(ndata):
                    sensmat[i, j] = thislam(xdata[i])

            sigma = float(data['ssr']) / (float(ndata) - float(nlinterms))
            ci = np.zeros([nlinterms])
            covar = sigma * np.linalg.inv(np.matmul(np.transpose(sensmat), sensmat))
            for i in range(0, nlinterms):
                ci[i] = t.ppf(1 - 0.025, int(ndata) - nlinterms) * np.sqrt(covar[i][i])
 
            data['covariance'][okey] = covar
            data['conf_inv'][okey] = list()
            for j in range(nlinterms):
                data['conf_inv'][okey].append('B' + str(j + 1) + ' : ' 
                                              + str(coeffs[j]) + '+/-' + str(ci[j]))
    else:
        model = data['model']
        model.split('=')[1]
        # out = model.split('=')[0]
        model = model.split('=')[1]
        # split the model on +/- to isolate each linear term
        # This section is not currently in compliance with custom basis functions
        model = model.replace(' - ', ' + ').split(" + ")
        while ('' in model): 
            model.remove('')
        while (' ' in model): 
            model.remove(' ')
        nlinterms = len(model)
        sensmat = np.zeros([ndata, nlinterms])
        covar = np.zeros([nlinterms, nlinterms])
        coeffs = np.zeros([nlinterms])
        for j in range(nlinterms):
            if(' * ' in model[j]):
                thisterm = model[j].split(' * ')
                coeffs[j] = float(eval(thisterm[0]))
                thisterm = thisterm[-1]
            else:
                thisterm = model[j]
                coeffs[j] = float(eval(thisterm))
            thislam = lambdify([symbols(data['xlabels'])], 
                               parse_expr(thisterm.replace('^', '**')), "numpy")
            for i in range(ndata):
                sensmat[i, j] = thislam(xdata[i])
            
        sigma = float(data['ssr']) / (float(ndata) - float(nlinterms))

        covar = sigma * np.linalg.inv(np.matmul(np.transpose(sensmat), sensmat))
        ci = np.zeros([nlinterms])
        for i in range(0, nlinterms):
            ci[i] = t.ppf(1 - 0.025, int(ndata) - nlinterms) \
                * np.sqrt(covar[i][i])  # 2.7

        data['covariance'] = covar
        data['conf_inv'] = list()
        for j in range(nlinterms):
            data['conf_inv'].append('B' + str(j + 1)
                                    + ' : ' + str(coeffs[j]) + '+/-' + str(ci[j]))
        return data



# coding: utf-8
""" __init__.py for idaes module

Set up logging for the idaes module, and import plugins.
"""
import os
import logging.config
import pyomo.common.plugin
import idaes.config
import toml

from .ver import __version__  # noqa

_log = logging.getLogger(__name__)

# Create the general IDAES configuration block
_config = config.new_idaes_config_block()

# Standard locations for config file, binary libraries and executables, ...
try:
    if os.name == 'nt':  # Windows
        data_directory = os.path.join(os.environ['LOCALAPPDATA'], "idaes")
    else:  # any other OS
        data_directory = os.path.join(os.environ['HOME'], ".idaes")
except AttributeError:
    data_directory = None

# Standard location for executable binaries.
if data_directory is not None:
    bin_directory = os.path.join(data_directory, "bin")
else:
    bin_directory = None

# Standard location for IDAES library files.
if data_directory is not None:
    lib_directory = os.path.join(data_directory, "lib")
else:
    lib_directory = None

def _create_data_dir():
    """Create the IDAES directory to store data files in."""
    config.create_dir(data_directory)

def _create_bin_dir():
    """Create the IDAES directory to store executable files in."""
    _create_data_dir()
    config.create_dir(bin_directory)

def _create_lib_dir():
    """Create the IDAES directory to store library files in."""
    _create_data_dir()
    config.create_dir(lib_directory)

# Set default configuration.  Used TOML string to serve as an example for
# and definitive guide for IDAES configuration files.
config.read_config(toml.loads(config.default_config), _config)

# Try to read the global IDAES config file.
# Set where to look for config files
_global_config_file = os.path.join(data_directory, "idaes.conf")
_local_config_file = "idaes.conf"

# Try to read global config then local
config.read_config(_global_config_file, _config)
config.read_config(_local_config_file, _config)
_log.debug("'idaes' logger debug test")

if _config.use_idaes_solvers:
    # Add IDAES stuff to the path unless you configure otherwise
    os.environ['PATH'] = os.pathsep.join([bin_directory, os.environ['PATH']])
    if os.name == 'nt':  # Windows (this is to find MinGW libs)
        os.environ['PATH'] = os.pathsep.join([os.environ['PATH'], lib_directory])
    else: # Linux and OSX, so far no need for this, but maybe in future
        __orig_ld = os.environ.get('LD_LIBRARY_PATH', '')
        os.environ['LD_LIBRARY_PATH'] = os.pathsep.join(
            [__orig_ld, lib_directory])

# Load plugins, could read a config file later by calling _read_config, but
# plugins only automatiaclly import when 'idaes' is imported. Could call
# _import_packages again later though

# This make "idaes" the current plugin environment while importing plugins here
pyomo.common.plugin.push("idaes")  # Add idaes plugin environment at top of stack
# Import plugins standard IDAES plugins, non-optional plugins
config.import_packages(_config["plugins"]["required"], optional=False)
# Import contrib plugins, failure to import these is non-fatal.
config.import_packages(_config["plugins"]["optional"], optional=True)
# go back to the previous plugin environment (what it was before pushing "idaes")
pyomo.common.plugin.pop()



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
import pytest

from idaes.core import FlowsheetBlock
from idaes.property_models.cubic_eos.cubic_prop_pack import \
    cubic_roots_available
from idaes.property_models.cubic_eos import BT_PR

from pyomo.environ import (ConcreteModel,
                           Objective,
                           SolverFactory,
                           TerminationCondition,
                           value)


# Set module level pyest marker
pytestmark = pytest.mark.cubic_root
prop_available = cubic_roots_available()


@pytest.mark.skipif(not prop_available,
                    reason="Cubic root finder not available")
def test_T_sweep():
    m = ConcreteModel()

    m.fs = FlowsheetBlock(default={'dynamic': False})

    m.fs.props = BT_PR.BTParameterBlock(
            default={'valid_phase': ('Vap', 'Liq')})

    m.fs.state = m.fs.props.state_block_class(
            default={'parameters': m.fs.props,
                     'defined_state': True})

    m.fs.obj = Objective(expr=(m.fs.state.temperature - 510)**2)

    for logP in range(8, 13, 1):
        m.fs.obj.deactivate()

        m.fs.state.flow_mol.fix(100)
        m.fs.state.mole_frac_comp["benzene"].fix(0.5)
        m.fs.state.mole_frac_comp["toluene"].fix(0.5)
        m.fs.state.temperature.fix(300)
        m.fs.state.pressure.fix(10**(0.5*logP))

        m.fs.state.initialize(outlvl=0)

        m.fs.state.temperature.unfix()
        m.fs.obj.activate()

        solver = SolverFactory('ipopt')
        results = solver.solve(m, tee=True)

        assert results.solver.termination_condition == \
            TerminationCondition.optimal
        assert m.fs.state.flow_mol_phase["Liq"].value <= 1e-5


@pytest.mark.skipif(not prop_available,
                    reason="Cubic root finder not available")
def test_P_sweep():
    m = ConcreteModel()

    m.fs = FlowsheetBlock(default={'dynamic': False})

    m.fs.props = BT_PR.BTParameterBlock(
            default={'valid_phase': ('Vap', 'Liq')})

    m.fs.state = m.fs.props.state_block_class(
            default={'parameters': m.fs.props,
                     'defined_state': True})

    for T in range(370, 500, 25):
        m.fs.state.flow_mol.fix(100)
        m.fs.state.mole_frac_comp["benzene"].fix(0.5)
        m.fs.state.mole_frac_comp["toluene"].fix(0.5)
        m.fs.state.temperature.fix(T)
        m.fs.state.pressure.fix(1e5)

        m.fs.state.initialize(outlvl=0)

        solver = SolverFactory('ipopt')
        results = solver.solve(m)

        assert results.solver.termination_condition == \
            TerminationCondition.optimal

        while m.fs.state.pressure.value <= 1e6:
            m.fs.state.pressure.value = m.fs.state.pressure.value + 1e5
            solver = SolverFactory('ipopt')
            results = solver.solve(m)
            assert results.solver.termination_condition == \
                TerminationCondition.optimal
            print(T, m.fs.state.pressure.value)


@pytest.mark.skipif(not prop_available,
                    reason="Cubic root finder not available")
def test_T350_P1_x5():
    m = ConcreteModel()

    m.fs = FlowsheetBlock(default={'dynamic': False})

    m.fs.props = BT_PR.BTParameterBlock(
            default={'valid_phase': ('Vap', 'Liq')})

    m.fs.state = m.fs.props.state_block_class(
            default={'parameters': m.fs.props,
                     'defined_state': True})

    m.fs.state.flow_mol.fix(100)
    m.fs.state.mole_frac_comp["benzene"].fix(0.5)
    m.fs.state.mole_frac_comp["toluene"].fix(0.5)
    m.fs.state.temperature.fix(350)
    m.fs.state.pressure.fix(1e5)

    # Trigger build of enthalpy and entropy
    m.fs.state.enth_mol_phase
    m.fs.state.entr_mol_phase

    m.fs.state.initialize(outlvl=0)

    solver = SolverFactory('ipopt')
    solver.solve(m)

    assert pytest.approx(value(m.fs.state._teq), abs=1e-1) == 365
    assert pytest.approx(
            value(m.fs.state.compress_fact_phase["Liq"]), 1e-5) == 0.0035346
    assert pytest.approx(
            value(m.fs.state.compress_fact_phase["Vap"]), 1e-5) == 0.966749
    assert pytest.approx(
            value(m.fs.state.fug_coeff_phase_comp["Liq", "benzene"]),
            1e-5) == 0.894676
    assert pytest.approx(
            value(m.fs.state.fug_coeff_phase_comp["Liq", "toluene"]),
            1e-5) == 0.347566
    assert pytest.approx(
            value(m.fs.state.fug_coeff_phase_comp["Vap", "benzene"]),
            1e-5) == 0.971072
    assert pytest.approx(
            value(m.fs.state.fug_coeff_phase_comp["Vap", "toluene"]),
            1e-5) == 0.959791

    assert pytest.approx(
            value(m.fs.state.mole_frac_phase_comp["Liq", "benzene"]),
            1e-5) == 0.5
    assert pytest.approx(
            value(m.fs.state.mole_frac_phase_comp["Liq", "toluene"]),
            1e-5) == 0.5
    assert pytest.approx(
            value(m.fs.state.mole_frac_phase_comp["Vap", "benzene"]),
            1e-5) == 0.70584
    assert pytest.approx(
            value(m.fs.state.mole_frac_phase_comp["Vap", "toluene"]),
            1e-5) == 0.29416

    assert pytest.approx(
            value(m.fs.state.enth_mol_phase["Liq"]), 1e-5) == 38942.8
    assert pytest.approx(
            value(m.fs.state.enth_mol_phase["Vap"]), 1e-5) == 78048.7
    assert pytest.approx(
            value(m.fs.state.entr_mol_phase["Liq"]), 1e-5) == -367.558
    assert pytest.approx(
            value(m.fs.state.entr_mol_phase["Vap"]), 1e-5) == -269.0553


@pytest.mark.skipif(not prop_available,
                    reason="Cubic root finder not available")
def test_T350_P5_x5():
    m = ConcreteModel()

    m.fs = FlowsheetBlock(default={'dynamic': False})

    m.fs.props = BT_PR.BTParameterBlock(
            default={'valid_phase': ('Vap', 'Liq')})

    m.fs.state = m.fs.props.state_block_class(
            default={'parameters': m.fs.props,
                     'defined_state': True})

    m.fs.state.flow_mol.fix(100)
    m.fs.state.mole_frac_comp["benzene"].fix(0.5)
    m.fs.state.mole_frac_comp["toluene"].fix(0.5)
    m.fs.state.temperature.fix(350)
    m.fs.state.pressure.fix(5e5)

    # Trigger build of enthalpy and entropy
    m.fs.state.enth_mol_phase
    m.fs.state.entr_mol_phase

    m.fs.state.initialize(outlvl=0)

    solver = SolverFactory('ipopt')
    solver.solve(m)

    assert pytest.approx(value(m.fs.state._teq), 1e-5) == 431.47
    assert pytest.approx(
            value(m.fs.state.compress_fact_phase["Liq"]), 1e-5) == 0.01766
    assert pytest.approx(
            value(m.fs.state.compress_fact_phase["Vap"]), 1e-5) == 0.80245
    assert pytest.approx(
            value(m.fs.state.fug_coeff_phase_comp["Liq", "benzene"]),
            1e-5) == 0.181229
    assert pytest.approx(
            value(m.fs.state.fug_coeff_phase_comp["Liq", "toluene"]),
            1e-5) == 0.070601
    assert pytest.approx(
            value(m.fs.state.fug_coeff_phase_comp["Vap", "benzene"]),
            1e-5) == 0.856523
    assert pytest.approx(
            value(m.fs.state.fug_coeff_phase_comp["Vap", "toluene"]),
            1e-5) == 0.799237

    assert pytest.approx(
            value(m.fs.state.mole_frac_phase_comp["Liq", "benzene"]),
            1e-5) == 0.5
    assert pytest.approx(
            value(m.fs.state.mole_frac_phase_comp["Liq", "toluene"]),
            1e-5) == 0.5
    assert pytest.approx(
            value(m.fs.state.mole_frac_phase_comp["Vap", "benzene"]),
            1e-5) == 0.65415
    assert pytest.approx(
            value(m.fs.state.mole_frac_phase_comp["Vap", "toluene"]),
            1e-5) == 0.34585

    assert pytest.approx(
            value(m.fs.state.enth_mol_phase["Liq"]), 1e-5) == 38966.9
    assert pytest.approx(
            value(m.fs.state.enth_mol_phase["Vap"]), 1e-5) == 75150.7
    assert pytest.approx(
            value(m.fs.state.entr_mol_phase["Liq"]), 1e-5) == -367.6064
    assert pytest.approx(
            value(m.fs.state.entr_mol_phase["Vap"]), 1e-5) == -287.3318


@pytest.mark.skipif(not prop_available,
                    reason="Cubic root finder not available")
def test_T450_P1_x5():
    m = ConcreteModel()

    m.fs = FlowsheetBlock(default={'dynamic': False})

    m.fs.props = BT_PR.BTParameterBlock(
            default={'valid_phase': ('Vap', 'Liq')})

    m.fs.state = m.fs.props.state_block_class(
            default={'parameters': m.fs.props,
                     'defined_state': True})

    m.fs.state.flow_mol.fix(100)
    m.fs.state.mole_frac_comp["benzene"].fix(0.5)
    m.fs.state.mole_frac_comp["toluene"].fix(0.5)
    m.fs.state.temperature.fix(450)
    m.fs.state.pressure.fix(1e5)

    # Trigger build of enthalpy and entropy
    m.fs.state.enth_mol_phase
    m.fs.state.entr_mol_phase

    m.fs.state.initialize(outlvl=0)

    solver = SolverFactory('ipopt')
    solver.solve(m)

    assert pytest.approx(value(m.fs.state._teq), 1e-5) == 371.4
    assert pytest.approx(
            value(m.fs.state.compress_fact_phase["Liq"]), 1e-5) == 0.0033583
    assert pytest.approx(
            value(m.fs.state.compress_fact_phase["Vap"]), 1e-5) == 0.9821368
    assert pytest.approx(
            value(m.fs.state.fug_coeff_phase_comp["Liq", "benzene"]),
            1e-5) == 8.069323
    assert pytest.approx(
            value(m.fs.state.fug_coeff_phase_comp["Liq", "toluene"]),
            1e-5) == 4.304955
    assert pytest.approx(
            value(m.fs.state.fug_coeff_phase_comp["Vap", "benzene"]),
            1e-5) == 0.985365
    assert pytest.approx(
            value(m.fs.state.fug_coeff_phase_comp["Vap", "toluene"]),
            1e-5) == 0.979457

    assert pytest.approx(
            value(m.fs.state.mole_frac_phase_comp["Liq", "benzene"]),
            1e-5) == 0.29861
    assert pytest.approx(
            value(m.fs.state.mole_frac_phase_comp["Liq", "toluene"]),
            1e-5) == 0.70139
    assert pytest.approx(
            value(m.fs.state.mole_frac_phase_comp["Vap", "benzene"]),
            1e-5) == 0.5
    assert pytest.approx(
            value(m.fs.state.mole_frac_phase_comp["Vap", "toluene"]),
            1e-5) == 0.5

    assert pytest.approx(
            value(m.fs.state.enth_mol_phase["Liq"]), 1e-5) == 49441.2
    assert pytest.approx(
            value(m.fs.state.enth_mol_phase["Vap"]), 1e-5) == 84175.1
    assert pytest.approx(
            value(m.fs.state.entr_mol_phase["Liq"]), 1e-5) == -333.836
    assert pytest.approx(
            value(m.fs.state.entr_mol_phase["Vap"]), 1e-5) == -247.385


@pytest.mark.skipif(not prop_available,
                    reason="Cubic root finder not available")
def test_T450_P5_x5():
    m = ConcreteModel()

    m.fs = FlowsheetBlock(default={'dynamic': False})

    m.fs.props = BT_PR.BTParameterBlock(
            default={'valid_phase': ('Vap', 'Liq')})

    m.fs.state = m.fs.props.state_block_class(
            default={'parameters': m.fs.props,
                     'defined_state': True})

    m.fs.state.flow_mol.fix(100)
    m.fs.state.mole_frac_comp["benzene"].fix(0.5)
    m.fs.state.mole_frac_comp["toluene"].fix(0.5)
    m.fs.state.temperature.fix(450)
    m.fs.state.pressure.fix(5e5)

    # Trigger build of enthalpy and entropy
    m.fs.state.enth_mol_phase
    m.fs.state.entr_mol_phase

    m.fs.state.initialize(outlvl=0)

    solver = SolverFactory('ipopt')
    solver.solve(m)

    assert pytest.approx(value(m.fs.state._teq), 1e-5) == 436.93
    assert pytest.approx(
            value(m.fs.state.compress_fact_phase["Liq"]), 1e-5) == 0.0166181
    assert pytest.approx(
            value(m.fs.state.compress_fact_phase["Vap"]), 1e-5) == 0.9053766
    assert pytest.approx(
            value(m.fs.state.fug_coeff_phase_comp["Liq", "benzene"]),
            1e-5) == 1.63308
    assert pytest.approx(
            value(m.fs.state.fug_coeff_phase_comp["Liq", "toluene"]),
            1e-5) == 0.873213
    assert pytest.approx(
            value(m.fs.state.fug_coeff_phase_comp["Vap", "benzene"]),
            1e-5) == 0.927534
    assert pytest.approx(
            value(m.fs.state.fug_coeff_phase_comp["Vap", "toluene"]),
            1e-5) == 0.898324

    assert pytest.approx(
            value(m.fs.state.mole_frac_phase_comp["Liq", "benzene"]),
            1e-5) == 0.3488737
    assert pytest.approx(
            value(m.fs.state.mole_frac_phase_comp["Liq", "toluene"]),
            1e-5) == 0.6511263
    assert pytest.approx(
            value(m.fs.state.mole_frac_phase_comp["Vap", "benzene"]),
            1e-5) == 0.5
    assert pytest.approx(
            value(m.fs.state.mole_frac_phase_comp["Vap", "toluene"]),
            1e-5) == 0.5

    assert pytest.approx(
            value(m.fs.state.enth_mol_phase["Liq"]), 1e-5) == 51095.2
    assert pytest.approx(
            value(m.fs.state.enth_mol_phase["Vap"]), 1e-5) == 83362.3
    assert pytest.approx(
            value(m.fs.state.entr_mol_phase["Liq"]), 1e-5) == -331.676
    assert pytest.approx(
            value(m.fs.state.entr_mol_phase["Vap"]), 1e-5) == -261.961


@pytest.mark.skipif(not prop_available,
                    reason="Cubic root finder not available")
def test_T368_P1_x5():
    m = ConcreteModel()

    m.fs = FlowsheetBlock(default={'dynamic': False})

    m.fs.props = BT_PR.BTParameterBlock(
            default={'valid_phase': ('Vap', 'Liq')})

    m.fs.state = m.fs.props.state_block_class(
            default={'parameters': m.fs.props,
                     'defined_state': True})

    m.fs.state.flow_mol.fix(100)
    m.fs.state.mole_frac_comp["benzene"].fix(0.5)
    m.fs.state.mole_frac_comp["toluene"].fix(0.5)
    m.fs.state.temperature.fix(368)
    m.fs.state.pressure.fix(1e5)

    # Trigger build of enthalpy and entropy
    m.fs.state.enth_mol_phase
    m.fs.state.entr_mol_phase

    m.fs.state.initialize(outlvl=0)

    solver = SolverFactory('ipopt')
    solver.solve(m)

    assert pytest.approx(value(m.fs.state._teq), 1e-5) == 368
    assert pytest.approx(
            value(m.fs.state.compress_fact_phase["Liq"]), 1e-5) == 0.003504
    assert pytest.approx(
            value(m.fs.state.compress_fact_phase["Vap"]), 1e-5) == 0.97
    assert pytest.approx(
            value(m.fs.state.fug_coeff_phase_comp["Liq", "benzene"]),
            1e-5) == 1.492049
    assert pytest.approx(
            value(m.fs.state.fug_coeff_phase_comp["Liq", "toluene"]),
            1e-5) == 0.621563
    assert pytest.approx(
            value(m.fs.state.fug_coeff_phase_comp["Vap", "benzene"]),
            1e-5) == 0.97469
    assert pytest.approx(
            value(m.fs.state.fug_coeff_phase_comp["Vap", "toluene"]),
            1e-5) == 0.964642

    assert pytest.approx(
            value(m.fs.state.mole_frac_phase_comp["Liq", "benzene"]),
            1e-5) == 0.4012128
    assert pytest.approx(
            value(m.fs.state.mole_frac_phase_comp["Liq", "toluene"]),
            1e-5) == 0.5987872
    assert pytest.approx(
            value(m.fs.state.mole_frac_phase_comp["Vap", "benzene"]),
            1e-5) == 0.6141738
    assert pytest.approx(
            value(m.fs.state.mole_frac_phase_comp["Vap", "toluene"]),
            1e-5) == 0.3858262

    assert pytest.approx(
            value(m.fs.state.enth_mol_phase["Liq"]), 1e-5) == 38235.1
    assert pytest.approx(
            value(m.fs.state.enth_mol_phase["Vap"]), 1e-5) == 77155.4
    assert pytest.approx(
            value(m.fs.state.entr_mol_phase["Liq"]), 1e-5) == -364.856
    assert pytest.approx(
            value(m.fs.state.entr_mol_phase["Vap"]), 1e-5) == -267.892


@pytest.mark.skipif(not prop_available,
                    reason="Cubic root finder not available")
def test_T376_P1_x2():
    m = ConcreteModel()

    m.fs = FlowsheetBlock(default={'dynamic': False})

    m.fs.props = BT_PR.BTParameterBlock(
            default={'valid_phase': ('Vap', 'Liq')})

    m.fs.state = m.fs.props.state_block_class(
            default={'parameters': m.fs.props,
                     'defined_state': True})

    m.fs.state.flow_mol.fix(100)
    m.fs.state.mole_frac_comp["benzene"].fix(0.2)
    m.fs.state.mole_frac_comp["toluene"].fix(0.8)
    m.fs.state.temperature.fix(376)
    m.fs.state.pressure.fix(1e5)

    # Trigger build of enthalpy and entropy
    m.fs.state.enth_mol_phase
    m.fs.state.entr_mol_phase

    m.fs.state.initialize(outlvl=0)

    solver = SolverFactory('ipopt')
    solver.solve(m)

    assert pytest.approx(value(m.fs.state._teq), 1e-5) == 376
    assert pytest.approx(
            value(m.fs.state.compress_fact_phase["Liq"]), 1e-5) == 0.00361333
    assert pytest.approx(
            value(m.fs.state.compress_fact_phase["Vap"]), 1e-5) == 0.968749
    assert pytest.approx(
            value(m.fs.state.fug_coeff_phase_comp["Liq", "benzene"]),
            1e-5) == 1.8394188
    assert pytest.approx(
            value(m.fs.state.fug_coeff_phase_comp["Liq", "toluene"]),
            1e-5) == 0.7871415
    assert pytest.approx(
            value(m.fs.state.fug_coeff_phase_comp["Vap", "benzene"]),
            1e-5) == 0.9763608
    assert pytest.approx(
            value(m.fs.state.fug_coeff_phase_comp["Vap", "toluene"]),
            1e-5) == 0.9663611

    assert pytest.approx(
            value(m.fs.state.mole_frac_phase_comp["Liq", "benzene"]),
            1e-5) == 0.17342
    assert pytest.approx(
            value(m.fs.state.mole_frac_phase_comp["Liq", "toluene"]),
            1e-5) == 0.82658
    assert pytest.approx(
            value(m.fs.state.mole_frac_phase_comp["Vap", "benzene"]),
            1e-5) == 0.3267155
    assert pytest.approx(
            value(m.fs.state.mole_frac_phase_comp["Vap", "toluene"]),
            1e-5) == 0.6732845

    assert pytest.approx(
            value(m.fs.state.enth_mol_phase["Liq"]), 1e-5) == 31535.8
    assert pytest.approx(
            value(m.fs.state.enth_mol_phase["Vap"]), 1e-5) == 69175.3
    assert pytest.approx(
            value(m.fs.state.entr_mol_phase["Liq"]), 1e-5) == -372.869
    assert pytest.approx(
            value(m.fs.state.entr_mol_phase["Vap"]), 1e-5) == -278.766



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Base for IDAES process model objects.
"""

import sys
import logging
import textwrap

from pyomo.core.base.block import _BlockData
from pyomo.core.base.misc import tabular_writer
from pyomo.environ import Block, value
from pyomo.gdp import Disjunct
from pyomo.common.config import ConfigBlock
from pyutilib.enum import Enum

from idaes.core.process_block import declare_process_block_class
from idaes.core.util.exceptions import (ConfigurationError,
                                        DynamicError,
                                        PropertyPackageError)
from idaes.core.util.tables import stream_table_dataframe_to_string
from idaes.core.util.model_statistics import (degrees_of_freedom,
                                              number_variables,
                                              number_activated_constraints,
                                              number_activated_blocks)


# Some more inforation about this module
__author__ = "John Eslick, Qi Chen, Andrew Lee"


__all__ = ['ProcessBlockData']


useDefault = object()


# Set up logger
_log = logging.getLogger(__name__)


# Enumerate options for material flow basis
MaterialFlowBasis = Enum(
    'molar',
    'mass',
    'other')


@declare_process_block_class("ProcessBaseBlock")
class ProcessBlockData(_BlockData):
    """
    Base class for most IDAES process models and classes.

    The primary purpose of this class is to create the local config block to
    handle arguments provided by the user when constructing an object and to
    ensure that these arguments are stored in the config block.

    Additionally, this class contains a number of methods common to all IDAES
    classes.
    """
    CONFIG = ConfigBlock("ProcessBlockData", implicit=False)

    def __init__(self, component):
        """
        Initialize a ProcessBlockData object.

        Args:
            component(Block): container Block instance to which this _BlockData
                              belongs.

        Returns:
            (ProcessBlockData): A new instance
        """
        super(ProcessBlockData, self).__init__(component=component)
        self._pb_configured = False

    def build(self):
        """
        The build method is called by the default ProcessBlock rule.  If a rule
        is sepecified other than the default it is important to call
        ProcessBlockData's build method to put information from the "default"
        and "initialize" arguments to a ProcessBlock derived class into the
        BlockData object's ConfigBlock.

        The the build method should usually be overloaded in a subclass derived
        from ProcessBlockData. This method would generally add Pyomo components
        such as variables, expressions, and constraints to the object. It is
        important for build() methods implimented in derived classes to call
        build() from the super class.

        Args:
            None

        Returns:
            None
        """
        self._get_config_args()

    def flowsheet(self):
        """
        This method returns the components parent flowsheet object, i.e. the
        flowsheet component to which the model is attached. If the component
        has no parent flowsheet, the method returns None.

        Args:
            None

        Returns:
            Flowsheet object or None
        """
        parent = self.parent_block()

        while True:
            if parent is None:
                return None

            if hasattr(parent, 'is_flowsheet') and parent.is_flowsheet():
                return parent

            else:
                parent = parent.parent_block()

    def _get_config_args(self):
        """
        Get config arguments for this element and put them in the ConfigBlock
        """
        if self._pb_configured:
            return
        self._pb_configured = True
        idx_map = self.parent_component()._idx_map  # index map function
        try:
            idx = self.index()
        except:
            idx = None
        if idx_map is not None:
            idx = idx_map(idx)
        initialize = self.parent_component()._block_data_config_initialize
        if idx in initialize:
            kwargs = initialize[idx]
        else:
            kwargs = self.parent_component()._block_data_config_default
        self.config = self.CONFIG(kwargs)

    def fix_initial_conditions(self, state="steady-state"):
        """This method fixes the initial conditions for dynamic models.

        Args:
            state : initial state to use for simulation (default =
                    'steady-state')

        Returns :
            None
        """
        if state == 'steady-state':
            for obj in self.component_objects((Block, Disjunct),
                                              descend_into=True):
                # Try to fix material_accumulation @ first time point
                try:
                    obj.material_accumulation[
                            obj.flowsheet().config.time.first(), ...].fix(0.0)
                except AttributeError:
                    pass

                # Try to fix element_accumulation @ first time point
                try:
                    obj.element_accumulation[
                            obj.flowsheet().config.time.first(), ...].fix(0.0)
                except AttributeError:
                    pass

                # Try to fix enthalpy_accumulation @ first time point
                try:
                    obj.enthalpy_accumulation[
                            obj.flowsheet().config.time.first(), ...].fix(0.0)
                except AttributeError:
                    pass

        else:
            raise ValueError("Unrecognised value for argument 'state'. "
                             "Valid values are 'steady-state'.")

    def unfix_initial_conditions(self):
        """This method unfixed the initial conditions for dynamic models.

        Args:
            None

        Returns :
            None
        """
        for obj in self.component_objects(Block, descend_into=True):
            # Try to unfix material_accumulation @ first time point
            try:
                obj.material_accumulation[
                        obj.flowsheet().config.time.first(), ...].unfix()
            except AttributeError:
                pass

            # Try to fix element_accumulation @ first time point
            try:
                obj.element_accumulation[
                        obj.flowsheet().config.time.first(), ...].unfix()
            except AttributeError:
                pass

            # Try to fix enthalpy_accumulation @ first time point
            try:
                obj.enthalpy_accumulation[
                        obj.flowsheet().config.time.first(), ...].unfix()
            except AttributeError:
                pass

    def report(self, time_point=0, dof=False, ostream=None, prefix=""):

        time_point = float(time_point)

        if ostream is None:
            ostream = sys.stdout

        # Get DoF and model stats
        if dof:
            dof_stat = degrees_of_freedom(self)
            nv = number_variables(self)
            nc = number_activated_constraints(self)
            nb = number_activated_blocks(self)

        # Get components to report in performance section
        performance = self._get_performance_contents(time_point=time_point)

        # Get stream table
        stream_table = self._get_stream_table_contents(time_point=time_point)

        # Set model type output
        if hasattr(self, "is_flowsheet") and self.is_flowsheet:
            model_type = "Flowsheet"
        else:
            model_type = "Unit"

        # Write output
        max_str_length = 84
        tab = " "*4
        ostream.write("\n"+"="*max_str_length+"\n")

        lead_str = f"{prefix}{model_type} : {self.name}"
        trail_str = f"Time: {time_point}"
        mid_str = " "*(max_str_length-len(lead_str)-len(trail_str))
        ostream.write(lead_str+mid_str+trail_str)

        if dof:
            ostream.write("\n"+"="*max_str_length+"\n")
            ostream.write(f"{prefix}{tab}Local Degrees of Freedom: {dof_stat}")
            ostream.write('\n')
            ostream.write(f"{prefix}{tab}Total Variables: {nv}{tab}"
                          f"Activated Constraints: {nc}{tab}"
                          f"Activated Blocks: {nb}")

        if performance is not None:
            ostream.write("\n"+"-"*max_str_length+"\n")
            ostream.write(f"{prefix}{tab}Unit Performance")
            ostream.write("\n"*2)
            if "vars" in performance.keys() and len(performance["vars"]) > 0:
                ostream.write(f"{prefix}{tab}Variables: \n\n")

                tabular_writer(
                        ostream,
                        prefix+tab,
                        ((k, v) for k, v in performance["vars"].items()),
                        ("Value", "Fixed", "Bounds"),
                        lambda k, v: [
                                "{:#.5g}".format(value(v)),
                                v.fixed,
                                v.bounds])

            if "exprs" in performance.keys() and len(performance["exprs"]) > 0:
                ostream.write("\n")
                ostream.write(f"{prefix}{tab}Expressions: \n\n")

                tabular_writer(
                        ostream,
                        prefix+tab,
                        ((k, v) for k, v in performance["exprs"].items()),
                        ("Value",),
                        lambda k, v: [
                                "{:#.5g}".format(value(v))])

            if ("params" in performance.keys() and
                    len(performance["params"]) > 0):
                ostream.write("\n")
                ostream.write(f"{prefix}{tab}Parameters: \n\n")

                tabular_writer(
                        ostream,
                        prefix+tab,
                        ((k, v) for k, v in performance["params"].items()),
                        ("Value", "Mutable"),
                        lambda k, v: [value(v),
                                      not v.is_constant()])

        if stream_table is not None:
            ostream.write("\n"+"-"*max_str_length+"\n")
            ostream.write(f"{prefix}{tab}Stream Table")
            ostream.write('\n')
            ostream.write(
                    textwrap.indent(
                            stream_table_dataframe_to_string(stream_table),
                            prefix+tab))
        ostream.write("\n"+"="*max_str_length+"\n")

    def _get_performance_contents(self, time_point):
        return None

    def _get_stream_table_contents(self, time_point):
        return None

    def _setup_dynamics(self):
        """
        This method automates the setting of the dynamic flag and time domain
        for unit models.

        Performs the following:
         1) Determines if this is a top level flowsheet
         2) Gets dynamic flag from parent if not top level, or checks validity
            of argument provided
         3) Checks has_holdup flag if present and dynamic = True

        Args:
            None

        Returns:
            None
        """
        # Get parent object
        if hasattr(self.parent_block(), "config"):
            # Parent block has a config block, so use this
            parent = self.parent_block()
        else:
            # Use parent flowsheet
            try:
                parent = self.flowsheet()
            except ConfigurationError:
                raise DynamicError('{} has no parent flowsheet from which to '
                                   'get dynamic argument. Please provide a '
                                   'value for this argument when constructing '
                                   'the unit.'
                                   .format(self.name))

        # Check the dynamic flag, and retrieve if necessary
        if self.config.dynamic == useDefault:
            # Get flag from parent flowsheet
            try:
                self.config.dynamic = parent.config.dynamic
            except AttributeError:
                # No flowsheet, raise exception
                raise DynamicError('{} parent flowsheet has no dynamic '
                                   'argument. Please provide a '
                                   'value for this argument when constructing '
                                   'the unit.'
                                   .format(self.name))

        # Check for case when dynamic=True, but parent dynamic=False
        if (self.config.dynamic and not parent.config.dynamic):
            raise DynamicError('{} trying to declare a dynamic model within '
                               'a steady-state flowsheet. This is not '
                               'supported by the IDAES framework. Try '
                               'creating a dynamic flowsheet instead, and '
                               'declaring some models as steady-state.'
                               .format(self.name))

        # Set and validate has_holdup argument
        if self.config.has_holdup == useDefault:
            # Default to same value as dynamic flag
            self.config.has_holdup = self.config.dynamic
        elif self.config.has_holdup is False:
            if self.config.dynamic is True:
                # Dynamic model must have has_holdup = True
                raise ConfigurationError(
                            "{} invalid arguments for dynamic and has_holdup. "
                            "If dynamic = True, has_holdup must also be True "
                            "(was False)".format(self.name))

    def _get_property_package(self):
        """
        This method gathers the necessary information about the property
        package to be used in the control volume block.

        If a property package has not been provided by the user, the method
        searches up the model tree until it finds an object with the
        'default_property_package' attribute and uses this package for the
        control volume block.

        The method also gathers any default construction arguments specified
        for the property package and combines these with any arguments
        specified by the user for the control volume block (user specified
        arguments take priority over defaults).

        Args:
            None

        Returns:
            None
        """
        # Get property_package block if not provided in arguments
        parent = self.parent_block()
        if self.config.property_package == useDefault:
            # Try to get property_package from parent
            try:
                if parent.config.property_package in [None, useDefault]:
                    parent.config.property_package = \
                        self._get_default_prop_pack()

                self.config.property_package = parent.config.property_package
            except AttributeError:
                self.config.property_package = self._get_default_prop_pack()

        # Check for any flowsheet level build arguments
        for k in self.config.property_package.config.default_arguments:
            if k not in self.config.property_package_args:
                self.config.property_package_args[k] = \
                    self.config.property_package.config.default_arguments[k]

    def _get_default_prop_pack(self):
        """
        This method is used to find a default property package defined at the
        flowsheet level if a package is not provided as an argument when
        instantiating the control volume block.

        Args:
            None

        Returns:
            None
        """
        parent = self.flowsheet()
        while True:
            if parent is None:
                raise ConfigurationError(
                            '{} no property package provided and '
                            'no default defined by parent flowsheet(s).'
                            .format(self.name))
            elif parent.config.default_property_package is not None:
                _log.info('{} Using default property package'
                          .format(self.name))
                return parent.config.default_property_package

            parent = parent.flowsheet()

    def _get_indexing_sets(self):
        """
        This method collects all necessary indexing sets from property
        parameter block and makes references to these for use within the
        control volume block. Collected indexing sets are phase_list and
        component_list.

        Args:
            None

        Returns:
            None
        """
        # Check for phase list(s)
        if not hasattr(self.config.property_package, "phase_list"):
            raise PropertyPackageError(
                '{} property_package provided does not '
                'contain a phase_list. '
                'Please contact the developer of the property package.'
                .format(self.name))

        # Check for component list(s)
        if not hasattr(self.config.property_package, "component_list"):
            raise PropertyPackageError(
                '{} property_package provided does not '
                'contain a component_list. '
                'Please contact the developer of the property package.'
                .format(self.name))

    def _get_reaction_package(self):
        """
        This method gathers the necessary information about the reaction
        package to be used in the control volume block (if required).

        If a reaction package has been provided by the user, the method
        gathers any default construction arguments specified
        for the reaction package and combines these with any arguments
        specified by the user for the control volume block (user specified
        arguments take priority over defaults).

        Args:
            None

        Returns:
            None
        """
        if self.config.reaction_package is not None:
            # Check for any flowsheet level build arguments
            for k in self.config.reaction_package.config.default_arguments:
                if k not in self.config.reaction_package_args:
                    self.config.reaction_package_args[k] = \
                        self.config.reaction_package.config.default_arguments[k]

    def _get_phase_comp_list(self):
        """
        Method to collect phase-component list from property package.
        If property package does not define a phase-component list, then it is
        assumed that all components are present in all phases.

        Args:
            None

        Returns:
            phase_component_list
        """
        # Get phase component list(s)
        if hasattr(self.config.property_package, "phase_component_list"):
            phase_component_list =\
                self.config.property_package.phase_component_list
        else:
            # Otherwise assume all components in all phases
            phase_component_list = {}
            for p in self.config.property_package.phase_list:
                phase_component_list[p] = self.config.property_package.\
                    component_list
        return phase_component_list
